"""
monobit.formats.xlfd.bdf - Adobe Glyph Bitmap Distribution Format

(c) 2019--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from ...binary import ceildiv
from ...storage import loaders, savers
from ...magic import FileFormatError
from ...font import Font, Coord
from ...raster import Raster
from ...glyph import Glyph
from ...encoding import charmaps, NotFoundError
from ...taggers import tagmaps
from ...labels import Char, Codepoint, Tag

from .xlfd import parse_xlfd_properties, create_xlfd_properties
from .xlfd import create_xlfd_name, CUSTOM_PROP


@loaders.register(
    name='bdf',
    magic=(b'STARTFONT ',),
    patterns=('*.bdf',),
)
def load_bdf(instream):
    """
    Load font from Adobe Glyph Bitmap Distribution Format (BDF) file.
    """
    instream = instream.text
    comments, bdf_props, x_props = _read_bdf_global(instream)
    logging.info('bdf properties:')
    for name, value in bdf_props.items():
        logging.info('    %s: %s', name, value)
    logging.info('x properties:')
    for name, value in x_props.items():
        logging.info('    %s: %s', name, value)
    bdf_glyphs = _read_bdf_glyphs(instream)
    glyphs, properties = _convert_from_bdf(bdf_glyphs, bdf_props, x_props)
    font = Font(glyphs, comment=comments, **properties)
    # create char labels, if encoding recognised
    font = font.label()
    # store labels as char only if we're working in unicode
    if charmaps.is_unicode(font.encoding):
        font = font.label(codepoint_from=None)
    return font


@savers.register(linked=load_bdf)
def save_bdf(fonts, outstream):
    """
    Save font to Adobe Glyph Bitmap Distribution Format (BDF) file.
    """
    if len(fonts) > 1:
        raise FileFormatError('Can only save one font to BDF file.')
    # ensure codepoint values are set
    font = fonts[0]
    try:
        font = font.label(codepoint_from=font.encoding)
    except NotFoundError:
        pass
    _save_bdf(font, outstream.text)


##############################################################################
# BDF reader
# BDF specification: https://adobe-type-tools.github.io/font-tech-notes/pdfs/5005.BDF_Spec.pdf


def read_props(instream, ends, keep_end=False):
    """Read key-value properties with comments."""
    # read global section
    props = []
    comments = []
    keyword = ''
    for line in instream:
        line = line.strip()
        if not line:
            continue
        if line.startswith('COMMENT'):
            comments.append(line[8:])
            continue
        keyword, _, value = line.partition(' ')
        props.append((keyword, value))
        if keyword in ends:
            if not keep_end:
                del props[-1]
            break
        else:
            keyword = ''
    return props, comments, keyword


def _read_bdf_glyphs(instream):
    """Read character section."""
    # output
    bdf_glyphs = []
    for line in instream:
        line = line.rstrip()
        if not line:
            continue
        if line.startswith('ENDFONT'):
            break
        keyword, _, tag = line.partition(' ')
        if keyword != 'STARTCHAR':
            raise FileFormatError(f'Expected STARTCHAR, not {line}')
        glyph_props = {'STARTCHAR': tag}
        proplist, comments, _ = read_props(instream, ends=('BITMAP',))
        propdict = dict(proplist)
        glyph_props |= dict(
            DWIDTH=_bdf_ints(propdict.pop('DWIDTH', None)),
            SWIDTH=_bdf_ints(propdict.pop('SWIDTH', None)),
            VVECTOR=_bdf_ints(propdict.pop('VVECTOR', None)),
            DWIDTH1=_bdf_ints(propdict.pop('DWIDTH1', None)),
            SWIDTH1=_bdf_ints(propdict.pop('SWIDTH1', None)),
            BBX=_bdf_ints(propdict.pop('BBX', None)),
        )
        glyph_props |= propdict
        glyph_props['COMMENT'] = '\n'.join(comments)
        glyph_props = {_k: _v for _k, _v in glyph_props.items() if _v is not None}
        # convert from hex-string to raster
        width, height, _, _ = glyph_props['BBX']
        hexstr = ''.join(
            # remove excess bytes on each hex line
            instream.readline().strip()[:ceildiv(width, 8)*2]
            for _ in range(height)
        )
        try:
            raster = Raster.from_hex(hexstr, width, height)
        except ValueError as e:
            logging.warning(f'Could not read glyph `{tag}` {hexstr}: {e}')
        else:
            bdf_glyphs.append(glyph_props | {'raster': raster})
        line = instream.readline()
        if not line.startswith('ENDCHAR'):
            raise FileFormatError(f'Expected ENDCHAR, not {line}')
    return bdf_glyphs


def _read_bdf_global(instream):
    """Read global section of BDF file."""
    start_props, start_comments, end = read_props(instream, ends=('STARTPROPERTIES', 'CHARS'), keep_end=True)
    x_props, x_comments = {}, {}
    end_props, end_comments = {}, {}
    if end != 'CHARS':
        del start_props[-1]
        x_props, x_comments, end = read_props(instream, ends=('CHARS', 'ENDPROPERTIES'), keep_end=True)
        if end != 'CHARS':
            del x_props[-1]
            end_props, end_comments, _ = read_props(instream, ends=('CHARS',), keep_end=True)
    bdf_props = {**dict(start_props), **dict(end_props)}
    comments = [*start_comments, *x_comments, *end_comments]
    return '\n'.join(comments), bdf_props, dict(x_props)


##############################################################################
# converter

def _bdf_ints(instr):
    if instr is not None:
        return tuple(int(_p) for _p in instr.split())
    return None


def _convert_from_bdf(bdf_glyphs, bdf_props, x_props):
    """Convert BDF data to monobit glyphs and properties."""
    # parse meaningful metadata
    known, global_metrics, bdf_unparsed = _extract_known_bdf_properties(bdf_props)
    properties = _convert_bdf_properties(known)
    glyphs = _convert_bdf_glyphs(bdf_glyphs, global_metrics, known)
    xlfd_props = parse_xlfd_properties(x_props, known['FONT'])
    # consistency checks
    if known['NCHARS'] != len(bdf_glyphs):
        logging.warning('Number of characters found does not match CHARS declaration.')
    for key, value in bdf_unparsed.items():
        logging.warning(f'Unrecognised BDF property {key}={value}')
        # preserve as custom property namespace, avoid clashes with yaff props
        properties[f'{CUSTOM_PROP}.{key}'] = value
    for key, value in xlfd_props.items():
        if key in properties and properties[key] != value:
            logging.debug(
                'Inconsistency between BDF and XLFD properties: '
                '%s=%s (from XLFD) but %s=%s (from BDF). Taking BDF property.',
                key, value, key, properties[key]
            )
        else:
            properties[key] = value
    return glyphs, properties


def _extract_known_bdf_properties(bdf_props):
    """Extract and classify global BDF properties that we know and use."""
    known = dict(
        SIZE=_bdf_ints(bdf_props.pop('SIZE')),
        STARTFONT=bdf_props.pop('STARTFONT'),
        CONTENTVERSION=bdf_props.pop('CONTENTVERSION', None),
        NCHARS=int(bdf_props.pop('CHARS')),
        FONT=bdf_props.pop('FONT'),
        METRICSSET=int(bdf_props.pop('METRICSSET', '0')),
    )
    # we're not type converting global metrics
    # because we still need to override with (unconverted) glyph metrics
    global_metrics = dict(
        # global DWIDTH; use bounding box as fallback if not specified
        DWIDTH=_bdf_ints(
            bdf_props.pop(
                'DWIDTH', ' '.join(bdf_props['FONTBOUNDINGBOX'].split()[:2])
            )
        ),
        SWIDTH=_bdf_ints(bdf_props.pop('SWIDTH', '0 0')),
        VVECTOR=_bdf_ints(bdf_props.pop('VVECTOR', '0 0')),
        DWIDTH1=_bdf_ints(bdf_props.pop('DWIDTH1', '0 0')),
        SWIDTH1=_bdf_ints(bdf_props.pop('SWIDTH1', '0 0')),
        BBX=_bdf_ints(bdf_props.pop('FONTBOUNDINGBOX')),
    )
    # keep unparsed bdf props
    return known, global_metrics, bdf_props


def _convert_bdf_properties(bdf_props):
    """Convert BDF global properties."""
    size, xdpi, ydpi, *depth_info = bdf_props['SIZE']
    if depth_info and depth_info[0] != 1:
        # Microsoft greymap extension of BDF, FontForge "BDF 2.3"
        # https://fontforge.org/docs/techref/BDFGrey.html
        raise FileFormatError('Greymap BDF not supported.')
    properties = {
        'source_format': f"BDF v{bdf_props['STARTFONT']}",
        'point_size': size,
        'dpi': (xdpi, ydpi),
        'revision': bdf_props['CONTENTVERSION'],
    }
    return properties


def _convert_bdf_labels(props):
    """Convert BDF glyph tags and encoding values to monobit labels."""
    labels = []
    # store STARTCHAR labels, if they're not just ordinals
    tag = props['STARTCHAR']
    try:
        int(tag)
    except ValueError:
        labels.append(Tag(tag))
    # ENCODING must be single integer or -1 followed by integer
    *_, encvalue = _bdf_ints(props['ENCODING'])
    if encvalue >= 0:
        labels.append(Codepoint(encvalue))
    return labels


def _convert_bdf_glyphs(bdf_glyphs, global_metrics, bdf_props):
    """Convert glyph properties."""
    if bdf_props['METRICSSET'] not in (0, 1, 2):
        logging.warning(
            f"Unsupported value METRICSSET={bdf_props['METRICSSET']} ignored"
        )
    glyphs = []
    for props in bdf_glyphs:
        raster = props.pop('raster')
        # fall back to glabal metrics, if not defined per-glyph
        props = global_metrics | props
        new_props = {}
        if bdf_props['METRICSSET'] != 1:
            new_props.update(_convert_horiz_metrics(raster.width, props, bdf_props))
        if bdf_props['METRICSSET'] in (1, 2):
            new_props.update(_convert_vert_metrics(raster.height, props, bdf_props))
        labels = _convert_bdf_labels(props)
        glyphs.append(Glyph(
            raster, labels=labels, comment=props['COMMENT'], **new_props
        ))
    return glyphs


def _convert_horiz_metrics(glyph_width, props, bdf_props):
    """Convert glyph horizontal metrics."""
    new_props = {}
    # bounding box & offset
    _bbx_width, _bbx_height, bboffx, shift_up = props['BBX']
    new_props['shift_up'] = shift_up
    # advance width
    dwidth_x, dwidth_y = props['DWIDTH']
    if dwidth_x > 0:
        advance_width = dwidth_x
        left_bearing = bboffx
    else:
        advance_width = -dwidth_x
        # bboffx would likely be negative
        left_bearing = advance_width + bboffx
    new_props['left_bearing'] = left_bearing
    new_props['right_bearing'] = advance_width - glyph_width - left_bearing
    # scalable width
    swidth_x, swidth_y = props['SWIDTH']
    new_props['scalable_width'] = swidth_to_pixel(
        swidth_x, point_size=bdf_props['SIZE'][0], dpi=bdf_props['SIZE'][1]
    )
    if new_props['scalable_width'] == advance_width:
        new_props['scalable_width'] = None
    if dwidth_y or swidth_y:
        logging.warning(
            'Vertical advance in horizontal writing not supported; ignored'
        )
    return new_props


def _convert_vert_metrics(glyph_height, props, bdf_props):
    """Convert glyph vertical metrics."""
    new_props = {}
    # bounding box & offset
    bbx_width, _bbx_height, bboffx, bboffy = props['BBX']
    voffx, voffy = props['VVECTOR']
    to_bottom = bboffy - voffy
    # vector from baseline to raster left; negative: baseline to right of left raster edge
    to_left = bboffx - voffx
    # leftward shift from baseline to raster central axis
    new_props['shift_left'] = ceildiv(bbx_width, 2) + to_left
    # advance height
    dwidth1_x, dwidth1_y = props['DWIDTH1']
    # dwidth1 vector: negative is down
    if dwidth1_y < 0:
        advance_height = -dwidth1_y
        top_bearing = -to_bottom - glyph_height
        bottom_bearing = advance_height - glyph_height - top_bearing
    else:
        advance_height = dwidth1_y
        bottom_bearing = to_bottom
        top_bearing = advance_height - glyph_height - bottom_bearing
    new_props['top_bearing'] = top_bearing
    new_props['bottom_bearing'] = bottom_bearing
    swidth1_x, swidth1_y = props['SWIDTH1']
    new_props['scalable_height'] = swidth_to_pixel(
        swidth1_y, point_size=bdf_props['SIZE'][0], dpi=bdf_props['SIZE'][2]
    )
    if new_props['scalable_height'] == advance_height:
        new_props['scalable_height'] = None
    if dwidth1_x or swidth1_x:
        logging.warning(
            'Horizontal advance in vertical writing not supported; ignored'
        )
    return new_props


def swidth_to_pixel(swidth, point_size, dpi):
    """DWIDTH = SWIDTH * points/1000 * dpi / 72"""
    return swidth * (point_size / 1000) * (dpi / 72)


##############################################################################
# BDF writer

def pixel_to_swidth(dwidth, point_size, dpi):
    """SWIDTH = DWIDTH / ( points/1000 * dpi / 72 )"""
    return int(
        round(dwidth / (point_size / 1000) / (dpi / 72))
    )


def _save_bdf(font, outstream):
    """Write one font to X11 BDF 2.1."""
    # property table
    xlfd_props = create_xlfd_properties(font)
    xlfd_name = create_xlfd_name(xlfd_props)
    bdf_props = _convert_to_bdf_properties(font, xlfd_name)
    # minimize glyphs to ink-bounds (BBX) before storing, except "cell" fonts
    if font.spacing not in ('character-cell', 'multi-cell'):
        font = font.reduce()
    # ensure character labels exist if needed
    if charmaps.is_unicode(font.encoding):
        font = font.label(match_whitespace=False, match_graphical=False)
    glyphs = tuple(
        _convert_to_bdf_glyph(glyph, font)
        for glyph in font.glyphs
    )
    # write out
    for key, value in bdf_props:
        if value:
            outstream.write(f'{key} {value}\n')
    if xlfd_props:
        outstream.write(f'STARTPROPERTIES {len(xlfd_props)}\n')
        for key, value in xlfd_props.items():
            outstream.write(f'{key} {value}\n')
        outstream.write('ENDPROPERTIES\n')
    outstream.write(f'CHARS {len(glyphs)}\n')
    for glyph in glyphs:
        for key, value in glyph:
            outstream.write(f'{key} {value}\n')
        outstream.write('ENDCHAR\n')
    outstream.write('ENDFONT\n')


def _convert_to_bdf_properties(font, xlfd_name):
    bdf_props = [
        ('STARTFONT', '2.1'),
    ] + [
        ('COMMENT', _comment) for _comment in font.get_comment().splitlines()
    ] + [
        ('FONT', xlfd_name),
        ('SIZE', f'{font.point_size} {font.dpi.x} {font.dpi.y}'),
        (
            # per the example in the BDF spec,
            # the first two coordinates in FONTBOUNDINGBOX
            # are the font's ink-bounds
            'FONTBOUNDINGBOX', (
                f'{font.bounding_box.x} {font.bounding_box.y} '
                f'{font.ink_bounds.left} {font.ink_bounds.bottom}'
            )
        )
    ]
    if font.has_vertical_metrics():
        bdf_props.append(('METRICSSET', '2'))
    return bdf_props


def _get_glyph_encvalue(glyph, is_unicode):
    """Get BDF ENCODING value and STARTCHAR tag."""
    if is_unicode:
        if len(glyph.char) == 1:
            encoding = ord(glyph.char)
        else:
            # multi-codepoint grapheme cluster or not set
            # -1 means no encoding value in bdf
            encoding = -1
    elif glyph.codepoint:
        # encoding values above 256 become multi-byte
        # unless we're working in unicode
        encoding = int(glyph.codepoint)
    else:
        encoding = -1
    # char must have a name in bdf
    for tag in glyph.tags:
        # bdf must only include printable ascii
        # postscript names (AGL) are purely alphanumeric
        # keep the first alphanumeric tag as the glyph name if available
        try:
            name = tag.value.encode('ascii').decode()
            if all(_c.isalnum() for _c in name):
                break
        except UnicodeError:
            pass
    else:
        # look up in adobe glyph list if character available
        name = tagmaps['adobe'].tag(*glyph.get_labels()).value
        # otherwise, use encoding value if available
        if not name and encoding != -1:
            name = f'char{encoding:02X}'
        if not name:
            logging.warning(
                f'Multi-codepoint glyph {glyph.codepoint}'
                "can't be stored as no name or character available."
            )
    return encoding, name


def _convert_to_bdf_glyph(glyph, font):
    encoding, name = _get_glyph_encvalue(
        glyph, charmaps.is_unicode(font.encoding)
    )
    swidth_y, dwidth_y = 0, 0
    # SWIDTH = DWIDTH / ( points/1000 * dpi / 72 )
    # DWIDTH specifies the widths in x and y, dwx0 and dwy0, in device pixels.
    # Like SWIDTH , this width information is a vector indicating the position of
    # the next glyph’s origin relative to the origin of this glyph.
    dwidth_x = glyph.advance_width
    swidth_x = pixel_to_swidth(
        glyph.scalable_width, font.point_size, font.dpi.x
    )
    glyphdata = [
        ('STARTCHAR', name),
        ('ENCODING', str(encoding)),
        # "The SWIDTH y value should always be zero for a standard X font."
        # "The DWIDTH y value should always be zero for a standard X font."
        ('SWIDTH', f'{swidth_x} 0'),
        ('DWIDTH', f'{dwidth_x} 0'),
        ('BBX', (
            f'{glyph.width} {glyph.height} '
            f'{glyph.left_bearing} {glyph.shift_up}'
        )),
    ]
    if font.has_vertical_metrics():
        to_left = glyph.shift_left - ceildiv(glyph.width, 2)
        to_bottom = -glyph.top_bearing - glyph.height
        voffx = glyph.left_bearing - to_left
        voffy = glyph.shift_up - to_bottom
        # dwidth1 vector: negative is down
        dwidth1_y = -glyph.advance_height
        swidth1_y = pixel_to_swidth(
            -glyph.scalable_height, font.point_size, font.dpi.y
        )
        glyphdata.extend([
            ('VVECTOR', f'{voffx} {voffy}'),
            ('SWIDTH1', f'0 {swidth1_y}'),
            ('DWIDTH1', f'0 {dwidth1_y}'),
        ])
    # bitmap
    if not glyph.height:
        glyphdata.append(('BITMAP', ''))
    else:
        hex = glyph.as_hex().upper()
        width = len(hex) // glyph.height
        split_hex = [
            hex[_offs:_offs+width]
            for _offs in range(0, len(hex), width)
        ]
        glyphdata.append(('BITMAP', '\n' + '\n'.join(split_hex)))
    return glyphdata



"""
monobit.glyphmap - glyph maps

(c) 2019--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

try:
    from PIL import Image
except ImportError:
    Image = None

from .properties import Props
from .raster import Raster
from .blocks import  blockstr


class GlyphMap:

    def __init__(self, map=()):
        self._map = list(map)
        self._turns = 0
        self._scale_x = 1
        self._scale_y = 1

    def __iter__(self):
        return iter(self._map)

    def __contains__(self, item):
        return item in self._map

    def turn(self, clockwise=1, *, anti=0):
        turns = clockwise - anti
        if turns%2:
            self._scale_x, self._scale_y = self._scale_y, self._scale_x
        self._turns += turns

    def stretch(self, scale_x=1, scale_y=1):
        self._scale_x *= scale_x
        self._scale_y *= scale_y

    def append_glyph(self, glyph, x, y, sheet=0):
        """Insert glyph in glyph map."""
        self._map.append(Props(glyph=glyph, x=x, y=y, sheet=sheet))

    def reorder(self, mapping):
        """Rearrange glyphs through index mapping."""
        self._map = [
            self._map[mapping[_i]]
            for _i in range(len(self._map))
        ]

    def get_bounds(self):
        """Get extreme coordinates across all sheets. Bounds are inclusive."""
        last = max(_entry.sheet for _entry in self._map)
        min_x = min(_entry.x for _entry in self._map)
        min_y = min(_entry.y for _entry in self._map)
        # inclusive bounds - we included the width/height of the glyphs
        # e.g. if I have a 2-pixel wide glyph at x=0, I need a 2-pixel image
        max_x = max(_entry.x + _entry.glyph.width for _entry in self._map)
        max_y = max(_entry.y + _entry.glyph.height for _entry in self._map)
        return last, min_x, min_y, max_x, max_y

    def to_canvas(self, sheet=0):
        """Convert one sheet of the glyph map to canvas."""
        # note that this gives the bounds across *all* sheets
        _, min_x, min_y, max_x, max_y = self.get_bounds()
        # no +1 as bounds are inclusive
        canvas = Canvas.blank(max_x - min_x, max_y - min_y)
        for entry in self._map:
            if entry.sheet == sheet:
                canvas.blit(
                    entry.glyph, entry.x - min_x, entry.y - min_y, operator=max
                )
        return canvas.stretch(self._scale_x, self._scale_y).turn(self._turns)

    def to_images(
            self, *, paper, ink, border, invert_y=False,
            transparent=True
        ):
        """Draw images based on shhets in glyph map."""
        if not Image:
            raise ImportError('Rendering to image requires PIL module.')
        paper, ink, border = 0, 255, 32
        last, min_x, min_y, max_x, max_y = self.get_bounds()
        # no +1 as bounds are inclusive
        width, height = max_x - min_x, max_y - min_y
        images = [Image.new('L', (width, height), border) for _ in range(last+1)]
        for entry in self._map:
            charimg = Image.new('L', (entry.glyph.width, entry.glyph.height))
            data = entry.glyph.as_bits(ink, paper)
            charimg.putdata(data)
            if invert_y:
                target = (entry.x, entry.y)
            else:
                # Image has ttb y coords, we have btt
                # our character origin is bottom left
                target = (entry.x-min_x, height-entry.glyph.height+min_y-entry.y)
            if transparent:
                mask = charimg
            else:
                mask = None
            images[entry.sheet].paste(charimg, target, mask)
        images = tuple(
            _im.resize((self._scale_x*_im.width, self._scale_y*_im.height))
                .rotate(-90 * self._turns, expand=True)
            for _im in images
        )
        return images

    def as_image(
            self, *,
            ink=(255, 255, 255), paper=(0, 0, 0), border=(0, 0, 0),
            sheet=0,
        ):
        """Convert glyph map to image."""
        images = self.to_images(ink=ink, paper=paper, border=border)
        return images[sheet]

    def as_text(
            self, *,
            ink='@', paper='.', border='.',
            start='', end='',
            sheet=0,
        ):
        """Convert glyph map to text."""
        canvas = self.to_canvas(sheet=sheet)
        return canvas.as_text(
            ink=ink, paper=paper, border=border, start=start, end=end
        )

    def as_blocks(self, resolution=(2, 2)):
        """Convert glyph map to a string of quadrant block characters."""
        canvas = self.to_canvas(sheet=0)
        return canvas.as_blocks(resolution)


class Canvas(Raster):
    """Mutable raster for glyph maps."""

    _inner = list
    _outer = list
    _0 = 0
    _1 = 1
    _itemtype = int

    @classmethod
    def blank(cls, width, height, fill=-1):
        """Create a canvas in background colour."""
        canvas = [[fill]*width for _ in range(height)]
        # setting 0 and 1 will make Raster init leave the input alone
        return cls(canvas, _0=0, _1=1)

    def blit(self, raster, grid_x, grid_y, operator):
        """
        Draw a matrix onto a canvas
        (leaving exising ink in place, depending on operator).
        """
        if not raster.width or not self.width:
            return self
        matrix = raster.as_matrix()
        for work_y in reversed(range(raster.height)):
            if 0 <= grid_y + work_y < self.height:
                row = self._pixels[self.height - (grid_y + work_y) - 1]
                for work_x, ink in enumerate(matrix[raster.height - work_y - 1]):
                    if 0 <= grid_x + work_x < self.width:
                        row[grid_x + work_x] = operator(ink, row[grid_x + work_x])
        return self

    def as_text(
            self, *,
            ink='@', paper='.', border='.',
            start='', end=''
        ):
        """Convert raster to text."""
        if not self.height:
            return ''
        colourdict = {-1: border, 0: paper, 1: ink}
        contents = '\n'.join(
            ''.join(colourdict[_pix] for _pix in _row)
            for _row in self._pixels
        )
        return blockstr(''.join((start, contents, end)))



"""
monobit.formats.xlfd.xlfd - X11 Logical Font Description

(c) 2019--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from ...encoding import charmaps
from ...labels import Char


# dot property namespace for unrecognised properties
CUSTOM_PROP = 'custom'


##############################################################################
# specification

# XLFD conventions: https://www.x.org/releases/X11R7.6/doc/xorg-docs/specs/XLFD/xlfd.html
# charset and property registry: https://github.com/freedesktop/xorg-docs/blob/master/registry

# https://github.com/freedesktop/xorg-docs/blob/master/registry
#
# 14. FONT CHARSET (REGISTRY AND ENCODING) NAMES
#
# See Sections 3.1.2.12 of the XLFD.  For ISO standards, the format
# will generally be: "ISO" + <standard-number> + "-" + <part-number>
#
# Name						Reference
# ----						---------
# "DEC"						[27]
# 	registry prefix
# "DEC.CNS11643.1986-2"				[53]
# 	CNS11643 2-plane using the encoding
# 	suggested in that standard
# "DEC.DTSCS.1990-2"				[54]
# 	DEC Taiwan Supplemental Character Set
# "fujitsu.u90x01.1991-0"				[87]
# "fujitsu.u90x03.1991-0"				[87]
# "GB2312.1980-0"					[39],[12]
# 	China (PRC) Hanzi, GL encoding
# "GB2312.1980-1"					[39]
# 	(deprecated)
# 	China (PRC) Hanzi, GR encoding
# "HP-Arabic8"					[36]
# 	HPARABIC8 8-bit character set
# "HP-East8"					[36]
# 	HPEAST8 8-bit character set
# "HP-Greek8"					[36]
# 	HPGREEK8 8-bit character set
# "HP-Hebrew8"					[36]
# 	HPHEBREW8 8-bit character set
# "HP-Japanese15"					[36]
# 	HPJAPAN15 15-bit characer set,
# 	modified from industry defacto
# 	standard Shift-JIS
# "HP-Kana8"					[36]
# 	HPKANA8 8-bit character set
# "HP-Korean15"					[36]
# 	HPKOREAN15 15-bit character set
# "HP-Roman8"					[36]
# 	HPROMAN8 8-bit character set
# "HP-SChinese15"					[36]
# 	HPSCHINA15 15-bit character set for
# 	support of Simplified Chinese
# "HP-TChinese15"					[36]
# 	HPTCHINA15 15-bit character set for
# 	support of Traditional Chinese
# "HP-Turkish8"					[36]
# 	HPTURKISH8 8-bit character set
# "IPSYS"						[59]
# 	registry prefix
# "IPSYS.IE-1"					[59]
# "ISO2022"<REG>"-"<ENC>				[44]
# "ISO646.1991-IRV"				[107]
# 	ISO 646 International Reference Version
# "ISO8859-1"					[15],[12]
# 	ISO Latin alphabet No. 1
# "ISO8859-2"					[15],[12]
# 	ISO Latin alphabet No. 2
# "ISO8859-3"					[15],[12]
# 	ISO Latin alphabet No. 3
# "ISO8859-4"					[15],[12]
# 	ISO Latin alphabet No. 4
# "ISO8859-5"					[15],[12]
# 	ISO Latin/Cyrillic alphabet
# "ISO8859-6"					[15],[12]
# 	ISO Latin/Arabic alphabet
# "ISO8859-7"					[15],[12]
# 	ISO Latin/Greek alphabet
# "ISO8859-8"					[15],[12]
# 	ISO Latin/Hebrew alphabet
# "ISO8859-9"					[15],[12]
# 	ISO Latin alphabet No. 5
# "ISO8859-10"					[15],[12]
# 	ISO Latin alphabet No. 6
# "ISO8859-13"					[15],[12]
# 	ISO Latin alphabet No. 7
# "ISO8859-14"					[15],[12]
# 	ISO Latin alphabet No. 8
# "ISO8859-15"					[15],[12]
# 	ISO Latin alphabet No. 9
# "ISO8859-16"					[15],[12]
# 	ISO Latin alphabet No. 10
# "FCD8859-15"					[7]
# 	(deprecated)
# 	ISO Latin alphabet No. 9, Final Committee Draft
# "ISO10646-1"					[133]
# 	Unicode Universal Multiple-Octet Coded Character Set
# "ISO10646-MES"					[133]
# 	(deprecated)
# 	Unicode Minimum European Subset
# "JISX0201.1976-0"				[38],[12]
# 	8-Bit Alphanumeric-Katakana Code
# "JISX0208.1983-0"				[40],[12]
# 	Japanese Graphic Character Set,
# 	GL encoding
# "JISX0208.1990-0"				[71]
# 	Japanese Graphic Character Set,
# 	GL encoding
# "JISX0208.1983-1"				[40]
# 	(deprecated)
# 	Japanese Graphic Character Set,
# 	GR encoding
# "JISX0212.1990-0"				[72]
# 	Supplementary Japanese Graphic Character Set,
# 	GL encoding
# "KOI8-R"					[119]
# 	Cyrillic alphabet
# "KSC5601.1987-0"				[41],[12]
# 	Korean Graphic Character Set,
# 	GL encoding
# "KSC5601.1987-1"				[41]
# 	(deprecated)
# 	Korean Graphic Character Set,
# 	GR encoding
# "omron_CNS11643-0"				[45]
# "omron_CNS11643-1"				[45]
# "omron_BIG5-0"					[45]
# "omron_BIG5-1"					[45]
# "wn.tamil.1993"					[103]


# https://www.freshports.org/x11-fonts/encodings/
#
# share/fonts/encodings/adobe-dingbats.enc.gz
# share/fonts/encodings/adobe-standard.enc.gz
# share/fonts/encodings/adobe-symbol.enc.gz
# share/fonts/encodings/armscii-8.enc.gz
# share/fonts/encodings/ascii-0.enc.gz
# share/fonts/encodings/dec-special.enc.gz
# share/fonts/encodings/encodings.dir
# share/fonts/encodings/ibm-cp437.enc.gz
# share/fonts/encodings/ibm-cp850.enc.gz
# share/fonts/encodings/ibm-cp852.enc.gz
# share/fonts/encodings/ibm-cp866.enc.gz
# share/fonts/encodings/iso8859-11.enc.gz
# share/fonts/encodings/iso8859-13.enc.gz
# share/fonts/encodings/iso8859-16.enc.gz
# share/fonts/encodings/iso8859-6.16.enc.gz
# share/fonts/encodings/iso8859-6.8x.enc.gz
# share/fonts/encodings/large/big5.eten-0.enc.gz
# share/fonts/encodings/large/big5hkscs-0.enc.gz
# share/fonts/encodings/large/cns11643-1.enc.gz
# share/fonts/encodings/large/cns11643-2.enc.gz
# share/fonts/encodings/large/cns11643-3.enc.gz
# share/fonts/encodings/large/encodings.dir
# share/fonts/encodings/large/gb18030-0.enc.gz
# share/fonts/encodings/large/gb18030.2000-0.enc.gz
# share/fonts/encodings/large/gb18030.2000-1.enc.gz
# share/fonts/encodings/large/gb2312.1980-0.enc.gz
# share/fonts/encodings/large/gbk-0.enc.gz
# share/fonts/encodings/large/jisx0201.1976-0.enc.gz
# share/fonts/encodings/large/jisx0208.1990-0.enc.gz
# share/fonts/encodings/large/jisx0212.1990-0.enc.gz
# share/fonts/encodings/large/ksc5601.1987-0.enc.gz
# share/fonts/encodings/large/ksc5601.1992-3.enc.gz
# share/fonts/encodings/large/sun.unicode.india-0.enc.gz
# share/fonts/encodings/microsoft-cp1250.enc.gz
# share/fonts/encodings/microsoft-cp1251.enc.gz
# share/fonts/encodings/microsoft-cp1252.enc.gz
# share/fonts/encodings/microsoft-cp1253.enc.gz
# share/fonts/encodings/microsoft-cp1254.enc.gz
# share/fonts/encodings/microsoft-cp1255.enc.gz
# share/fonts/encodings/microsoft-cp1256.enc.gz
# share/fonts/encodings/microsoft-cp1257.enc.gz
# share/fonts/encodings/microsoft-cp1258.enc.gz
# share/fonts/encodings/microsoft-win3.1.enc.gz
# share/fonts/encodings/mulearabic-0.enc.gz
# share/fonts/encodings/mulearabic-1.enc.gz
# share/fonts/encodings/mulearabic-2.enc.gz
# share/fonts/encodings/mulelao-1.enc.gz
# share/fonts/encodings/suneu-greek.enc.gz
# share/fonts/encodings/tcvn-0.enc.gz
# share/fonts/encodings/tis620-2.enc.gz
# share/fonts/encodings/viscii1.1-1.enc.gz

# https://x.org/releases/X11R7.7/doc/xorg-docs/fonts/fonts.html
# > Specifying an encoding value of
# > adobe-fontspecific for a Type 1 font disables the encoding mechanism. This is useful with symbol
# > and incorrectly encoded fonts (see Hints about using badly encoded fonts below).
#
# > In the case of Type 1 fonts, the font designer can specify a default encoding; this encoding is
# > requested by using the “adobe-fontspecific” encoding in the XLFD name.

# some more encodings mentioned here:
# https://www.x.org/releases/X11R6.9.0/doc/html/fonts3.html
# https://www.x.org/releases/X11R6.9.0/doc/html/fonts4.html
#
# ISO10646-1 = unicode
# ISO8859-1 (etc)  = latin-1
# KOI8-R (-U -RU -UNI -E)
# fontspecific-0
# microsoft-symbol
# apple-roman
# ibm-cp437
# adobe-standard
# adobe-dingbats
# adobe-symbol
# adobe-fontspecific
# microsoft-cp1252
# microsoft-win3.1
# jisx0208.1990-0
# jisx0201.1976-0
# big5.eten-0
# gb2312.1980-0

# X11 / BDF undefined
# these are mapped to "no encoding"
# "no encoding" is mapped to the first name provided here
_UNDEFINED_ENCODINGS = [
    'fontspecific-0',
    'adobe-fontspecific',
    'fontspecific',
]

# names to be used when writing bdf
_UNIX_ENCODINGS = {
    '': _UNDEFINED_ENCODINGS[0],
    'unicode': 'ISO10646-1',
    'ascii': 'ascii-0',

    'latin-1': 'ISO8859-1',
    'latin-2': 'ISO8859-2',
    'latin-3': 'ISO8859-3',
    'latin-4': 'ISO8859-4',
    'iso8859-5': 'ISO8859-5',
    'iso8859-6': 'ISO8859-6',
    'iso8859-7': 'ISO8859-7',
    'iso8859-8': 'ISO8859-8',
    'iso8859-9': 'ISO8859-9',
    'iso8859-10': 'ISO8859-10',
    'iso8859-11': 'ISO8859-11',
    'iso8859-13': 'ISO8859-13',
    'iso8859-14': 'ISO8859-14',
    'iso8859-15': 'ISO8859-15',
    'iso8859-16': 'ISO8859-16',

    'koi8-r': 'KOI8-R',
    'koi8-u': 'KOI8-U',
    'koi8-ru': 'KOI8-RU',
    'koi8-e': 'KOI8-E',
    'koi8-unified': 'KOI8-UNI',

    'mac-symbol': 'microsoft-symbol',
    'mac-roman': 'apple-roman',
    'cp437': 'ibm-cp437',
    'cp850': 'ibm-cp850',
    'cp852': 'ibm-cp852',
    'cp866': 'ibm-cp866',

    # adobe, dec encodings have standard names

    'windows-1250': 'microsoft-cp1250',
    'windows-1251': 'microsoft-cp1251',
    'windows-1252': 'microsoft-cp1252',
    'windows-1253': 'microsoft-cp1253',
    'windows-1254': 'microsoft-cp1254',
    'windows-1255': 'microsoft-cp1255',
    'windows-1256': 'microsoft-cp1256',
    'windows-1257': 'microsoft-cp1257',
    'windows-1258': 'microsoft-cp1258',
    'windows-3.1': 'microsoft-win3.1',

    'hp-roman8': 'HP-Roman8',
    'hp-greek8': 'HP-Greek8',
    'hp-thai8': 'HP-Thai8',
    'hp-turkish8': 'HP-Turkish8',

    'jis-x0201': 'jisx0201.1976-0',

    'big5-hkscs': 'big5hkscs-0',
    'windows-936': 'gbk-0',

    # johab
    'windows-1361': 'ksc5601.1992-3',
    # ksc5601.1987-0

    'tis-620': 'tis620-0',
    'viscii': 'viscii1.1-1',
    # tcvn-0 ?

    # big5.eten-0
    # gb2312.1980-0
    # cns11643-1
    # cns11643-2
    # cns11643-3
    # gb18030-0
    # gb18030.2000-0
    # gb18030.2000-1

    # jisx0208.1990-0
    # jisx0212.1990-0
}

_SLANT_MAP = {
    'R': 'roman',
    'I': 'italic',
    'O': 'oblique',
    'RO': 'reverse-oblique',
    'RI': 'reverse-italic',
    'OT': '', # 'other'
}

_SPACING_MAP = {
    'P': 'proportional',
    'M': 'monospace',
    'C': 'character-cell',
}

_SETWIDTH_MAP = {
    '0': '', # Undefined or unknown
    '10': 'ultra-condensed',
    '20': 'extra-condensed',
    '30': 'condensed', # Condensed, Narrow, Compressed, ...
    '40': 'semi-condensed',
    '50': 'medium', # Medium, Normal, Regular, ...
    '60': 'semi-expanded', # SemiExpanded, DemiExpanded, ...
    '70': 'expanded',
    '80': 'extra-expanded', # ExtraExpanded, Wide, ...
    '90': 'ultra-expanded',
}

_WEIGHT_MAP = {
    '0': '', # Undefined or unknown
    '10': 'thin', # UltraLight
    '20': 'extra-light',
    '30': 'light',
    '40': 'semi-light', # SemiLight, Book, ...
    '50': 'regular', # Medium, Normal, Regular,...
    '60': 'semi-bold', # SemiBold, DemiBold, ...
    '70': 'bold',
    '80': 'extra-bold', # ExtraBold, Heavy, ...
    '90': 'heavy', # UltraBold, Black, ...,
}

# fields of the xlfd font name
_XLFD_NAME_FIELDS = (
    # key name is unofficial but in widespread use
    'FONTNAME_REGISTRY',
    # official field name matches
    'FOUNDRY',
    'FAMILY_NAME',
    'WEIGHT_NAME',
    'SLANT',
    'SETWIDTH_NAME',
    'ADD_STYLE_NAME',
    'PIXEL_SIZE',
    'POINT_SIZE',
    'RESOLUTION_X',
    'RESOLUTION_Y',
    'SPACING',
    'AVERAGE_WIDTH',
    'CHARSET_REGISTRY',
    'CHARSET_ENCODING',
)

# unparsed xlfd properties, for reference
_XLFD_UNPARSED = {
    # average cap/lower width, in tenths of pixels, negative for rtl
    'AVG_CAPITAL_WIDTH',
    'AVG_LOWERCASE_WIDTH',
    # width of a quad (em) space. deprecated.
    'QUAD_WIDTH',
    # for boxing/voiding glyphs
    'STRIKEOUT_ASCENT',
    'STRIKEOUT_DESCENT',
    # the nominal posture angle of the typeface design, in 1/64 degrees, measured from the
    # glyph origin counterclockwise from the three o’clock position.
    'ITALIC_ANGLE',
    # the calculated weight of the font, computed as the ratio of capital stem width to CAP_HEIGHT,
    # in the range 0 to 1000, where 0 is the lightest weight.
    'WEIGHT',
    # the format of the font data as they are read from permanent storage by the current font source
    # 'Bitmap', 'Prebuilt', 'Type 1', 'TrueType', 'Speedo', 'F3'
    'FONT_TYPE',
    # the specific name of the rasterizer that has performed some rasterization operation
    #(such as scaling from out- lines) on this font.
    'RASTERIZER_NAME',
    # the formal or informal version of a font rasterizer.
    'RASTERIZER_VERSION',
    #'RAW_*',
    # axes of a polymorphic font
    'AXIS_NAMES',
    'AXIS_LIMITS',
    'AXIS_TYPES',
    # key name is not in the spec but in widespread use and in bdflib
    'FONTNAME_REGISTRY',
    'CHARSET_COLLECTIONS',
    'DEVICE_FONT_NAME',

    # for fonts with a transformation matrix
    'RAW_ASCENT',
    'RAW_AVERAGE_WIDTH',
    'RAW_AVG_CAPITAL_WIDTH',
    'RAW_AVG_LOWERCASE_WIDTH',
    'RAW_CAP_HEIGHT',
    'RAW_DESCENT',
    'RAW_END_SPACE',
    'RAW_FIGURE_WIDTH',
    'RAW_MAX_SPACE',
    'RAW_MIN_SPACE',
    'RAW_NORM_SPACE',
    'RAW_PIXEL_SIZE',
    'RAW_POINT_SIZE',
    'RAW_PIXELSIZE',
    'RAW_POINTSIZE',
    'RAW_QUAD_WIDTH',
    'RAW_SMALL_CAP_SIZE',
    'RAW_STRIKEOUT_ASCENT',
    'RAW_STRIKEOUT_DESCENT',
    'RAW_SUBSCRIPT_SIZE',
    'RAW_SUBSCRIPT_X',
    'RAW_SUBSCRIPT_Y',
    'RAW_SUPERSCRIPT_SIZE',
    'RAW_SUPERSCRIPT_X',
    'RAW_SUPERSCRIPT_Y',
    'RAW_UNDERLINE_POSITION',
    'RAW_UNDERLINE_THICKNESS',
    'RAW_X_HEIGHT',
}


def _parse_xlfd_name(xlfd_str):
    """Parse X logical font description font name string."""
    if not xlfd_str:
        # if no name at all that's intentional (used for HBF) - do not warn
        return {}
    xlfd = xlfd_str.split('-')
    if len(xlfd) == 15:
        properties = {
            _key: _value.replace('~', '-')
            for _key, _value in zip(_XLFD_NAME_FIELDS, xlfd)
            if _key and _value
        }
    else:
        logging.warning('Could not parse X font name string `%s`', xlfd_str)
        return {}
    return properties

def from_quoted_string(quoted):
    """Strip quotes"""
    return quoted.strip('"').replace('""', '"')

def _all_ints(*value, to_int=int):
    """Convert all items in tuple to int."""
    return tuple(to_int(_x) for _x in value)

def parse_xlfd_properties(x_props, xlfd_name, to_int=int):
    """Parse X metadata."""
    xlfd_name_props = _parse_xlfd_name(xlfd_name)
    # find fields in XLFD FontName that do not match the FontProperties
    conflicting = '\n'.join(
        f'{_k}={repr(_v)} vs {_k}={repr(x_props[_k])}' for _k, _v in xlfd_name_props.items()
        if _k in x_props and not (
            str(_v) == str(x_props[_k])
            or f'"{_v}"' == str(x_props[_k])
        )
    )
    if conflicting:
        logging.info('Conflicts between XLFD FontName and FontProperties: %s', conflicting)
    # continue with the XLFD FontProperties overriding the XLFD FontName fields if given
    x_props = {**xlfd_name_props, **x_props}
    # PIXEL_SIZE = ROUND((RESOLUTION_Y * POINT_SIZE) / 722.7)
    properties = {
        # FULL_NAME is deprecated
        'name': from_quoted_string(
            x_props.pop('FACE_NAME', x_props.pop('FULL_NAME', ''))
        ),
        'revision': from_quoted_string(x_props.pop('FONT_VERSION', '')),
        'foundry': from_quoted_string(x_props.pop('FOUNDRY', '')),
        'copyright': from_quoted_string(x_props.pop('COPYRIGHT', '')),
        'notice': from_quoted_string(x_props.pop('NOTICE', '')),
        'family': from_quoted_string(x_props.pop('FAMILY_NAME', '')),
        'style': from_quoted_string(x_props.pop('ADD_STYLE_NAME', '')).lower(),
        'ascent': x_props.pop('FONT_ASCENT', None),
        'descent': x_props.pop('FONT_DESCENT', None),
        'x_height': x_props.pop('X_HEIGHT', None),
        'cap_height': x_props.pop('CAP_HEIGHT', None),
        'pixel_size': x_props.pop('PIXEL_SIZE', None),
        'slant': _SLANT_MAP.get(
            from_quoted_string(x_props.pop('SLANT', '')), None
        ),
        'spacing': _SPACING_MAP.get(
            from_quoted_string(x_props.pop('SPACING', '')), None
        ),
        'underline_descent': x_props.pop('UNDERLINE_POSITION', None),
        'underline_thickness': x_props.pop('UNDERLINE_THICKNESS', None),
        'superscript_size': x_props.pop('SUPERSCRIPT_SIZE', None),
        'subscript_size': x_props.pop('SUBSCRIPT_SIZE', None),
        'small_cap_size': x_props.pop('SMALL_CAP_SIZE', None),
        'digit_width': x_props.pop('FIGURE_WIDTH', None),
        'min_word_space': x_props.pop('MIN_SPACE', None),
        'word_space': x_props.pop('NORM_SPACE', None),
        'max_word_space': x_props.pop('MAX_SPACE', None),
        'sentence_space': x_props.pop('END_SPACE', None),
    }
    if 'DESTINATION' in x_props and to_int(x_props['DESTINATION']) < 2:
        dest = to_int(x_props.pop('DESTINATION'))
        properties['device'] = 'screen' if dest else 'printer'
    if 'POINT_SIZE' in x_props:
        properties['point_size'] = round(to_int(x_props.pop('POINT_SIZE')) / 10)
    if 'AVERAGE_WIDTH' in x_props:
        properties['average_width'] = to_int(x_props.pop('AVERAGE_WIDTH')) / 10
    # prefer the more precise relative weight and setwidth measures
    if 'RELATIVE_SETWIDTH' in x_props:
        properties['setwidth'] = _SETWIDTH_MAP.get(
            x_props.pop('RELATIVE_SETWIDTH'), None
        )
        x_props.pop('SETWIDTH_NAME', None)
    if 'setwidth' not in properties or not properties['setwidth']:
        properties['setwidth'] = from_quoted_string(
            x_props.pop('SETWIDTH_NAME', '')
        ).lower()
    if 'RELATIVE_WEIGHT' in x_props:
        properties['weight'] = _WEIGHT_MAP.get(x_props.pop('RELATIVE_WEIGHT'), None)
        x_props.pop('WEIGHT_NAME', None)
    if 'weight' not in properties or not properties['weight']:
        properties['weight'] = from_quoted_string(x_props.pop('WEIGHT_NAME', '')).lower()
    # resolution
    if 'RESOLUTION_X' in x_props and 'RESOLUTION_Y' in x_props:
        properties['dpi'] = _all_ints(
            x_props.pop('RESOLUTION_X'), x_props.pop('RESOLUTION_Y')
        )
        x_props.pop('RESOLUTION', None)
    elif 'RESOLUTION' in x_props:
        # deprecated
        properties['dpi'] = _all_ints(
            x_props.get('RESOLUTION'), x_props.pop('RESOLUTION'), to_int=to_int
        )
    if 'SUPERSCRIPT_X' in x_props and 'SUPERSCRIPT_Y' in x_props:
        properties['superscript_offset'] = _all_ints(
            x_props.pop('SUPERSCRIPT_X'), x_props.pop('SUPERSCRIPT_Y'),
            to_int=to_int
        )
    if 'SUBSCRIPT_X' in x_props and 'SUBSCRIPT_Y' in x_props:
        properties['subscript_offset'] = _all_ints(
            x_props.pop('SUBSCRIPT_X'), x_props.pop('SUBSCRIPT_Y'),
            to_int=to_int
        )
    # encoding
    registry = from_quoted_string(x_props.pop('CHARSET_REGISTRY', '')).lower()
    encoding = from_quoted_string(x_props.pop('CHARSET_ENCODING', '')).lower()
    if registry and encoding and encoding != '0':
        properties['encoding'] = f'{registry}-{encoding}'
    elif registry:
        properties['encoding'] = registry
    elif encoding != '0':
        properties['encoding'] = encoding
    if properties['encoding'] in _UNDEFINED_ENCODINGS:
        properties['encoding'] = ''
    if 'DEFAULT_CHAR' in x_props:
        default_ord = to_int(x_props.pop('DEFAULT_CHAR', None))
        if charmaps.is_unicode(properties['encoding']):
            properties['default_char'] = Char(chr(default_ord))
        else:
            properties['default_char'] = default_ord
    # keep original FontName if invalid or conflicting
    if not xlfd_name_props:
        if not properties['family']:
            properties['family'] = xlfd_name
        else:
            properties['xlfd.font_name'] = xlfd_name
    if conflicting:
        properties['xlfd.font_name'] = xlfd_name
    # keep unparsed but known properties
    for key in _XLFD_UNPARSED:
        try:
            value = x_props.pop(key)
        except KeyError:
            continue
        key = key.lower()
        value = from_quoted_string(value)
        if value:
            properties[f'xlfd.{key}'] = value
    # drop empty known properties
    properties = {
        _k: _v for _k, _v in properties.items()
        if _v is not None and _v != ''
    }
    # keep unrecognised properties but in separate namespace
    # to avoid any clashes with yaff properties
    properties.update({
        f'{CUSTOM_PROP}.{_k}'.lower(): from_quoted_string(_v)
        for _k, _v in x_props.items()
    })
    return properties



##############################################################################

def create_xlfd_name(xlfd_props):
    """Construct XLFD name from properties."""
    # if we stored a font name explicitly, keep it
    try:
        return xlfd_props['FONT_NAME']
    except KeyError:
        pass
    xlfd_fields = [xlfd_props.get(prop, '') for prop in _XLFD_NAME_FIELDS]
    return '-'.join(str(_field).strip('"').replace('-', '~') for _field in xlfd_fields)

def _quoted_string(unquoted):
    """Return quoted version of string, if any."""
    return '"{}"'.format(unquoted.replace('"', '""'))

def create_xlfd_properties(font):
    """Construct XLFD properties."""
    # construct the fields needed for FontName if not defined, leave others optional
    xlfd_props = {
        'FONT_ASCENT': font.get_defined('ascent'),
        'FONT_DESCENT': font.get_defined('descent'),
        'PIXEL_SIZE': font.pixel_size,
        'X_HEIGHT': font.get_defined('x_height'),
        'CAP_HEIGHT': font.get_defined('cap_height'),
        'RESOLUTION_X': font.dpi.x,
        'RESOLUTION_Y': font.dpi.y,
        'POINT_SIZE': int(font.point_size) * 10,
        'FACE_NAME': _quoted_string(font.name) if 'name' in font.get_properties() else None,
        'FONT_VERSION': _quoted_string(font.revision) if 'revision' in font.get_properties() else None,
        'COPYRIGHT': _quoted_string(font.copyright) if 'copyright' in font.get_properties() else None,
        'NOTICE': _quoted_string(font.notice) if 'notice' in font.get_properties() else None,
        'FOUNDRY': _quoted_string(font.foundry),
        'FAMILY_NAME': _quoted_string(font.family),
        'WEIGHT_NAME': _quoted_string(font.weight.title()),
        'RELATIVE_WEIGHT': int(
            {_v: _k for _k, _v in _WEIGHT_MAP.items()}.get(font.weight, 50)
        ),
        'SLANT': _quoted_string(
            {_v: _k for _k, _v in _SLANT_MAP.items()}.get(font.slant, 'R')
        ),
        'SPACING': _quoted_string(
            {_v: _k for _k, _v in _SPACING_MAP.items()}.get(font.spacing, 'P')
        ),
        'SETWIDTH_NAME': _quoted_string(font.setwidth.title()),
        'RELATIVE_SETWIDTH': int(
            # 50 is medium
            {_v: _k for _k, _v in _SETWIDTH_MAP.items()}.get(font.setwidth, 50)
        ),
        'ADD_STYLE_NAME': _quoted_string(font.style.title()),
        'AVERAGE_WIDTH': round(float(font.average_width) * 10),
        # only set if explicitly defined
        'UNDERLINE_POSITION': font.get_defined('underline_descent'),
        'UNDERLINE_THICKNESS': font.get_defined('underline_thickness'),
        'DESTINATION': {'printer': 0, 'screen': 1}.get(font.device.lower(), None),
        'SUPERSCRIPT_SIZE': font.get_defined('superscript_size'),
        'SUBSCRIPT_SIZE': font.get_defined('subscript_size'),
        'SMALL_CAP_SIZE': font.get_defined('small_cap_size'),
        'SUPERSCRIPT_X': font.superscript_offset.x if 'superscript_offset' in font.get_properties() else None,
        'SUPERSCRIPT_Y': font.superscript_offset.y if 'superscript_offset' in font.get_properties() else None,
        'SUBSCRIPT_X': font.subscript_offset.x if 'subscript_offset' in font.get_properties() else None,
        'SUBSCRIPT_Y': font.subscript_offset.y if 'subscript_offset' in font.get_properties() else None,
        'FIGURE_WIDTH': font.get_defined('digit_width'),
        'MIN_SPACE': font.get_defined('min_word_space'),
        'NORM_SPACE': font.get_defined('word_space'),
        'MAX_SPACE': font.get_defined('max_word_space'),
        'END_SPACE': font.get_defined('sentence_space'),
    }
    # encoding dependent values
    default_glyph = font.get_default_glyph()
    if charmaps.is_unicode(font.encoding):
        if default_glyph.char:
            default_codepoint = tuple(ord(_c) for _c in default_glyph.char)
        else:
            default_codepoint = default_glyph.codepoint
        if not len(default_codepoint):
            logging.debug('Cannot store default glyph in BDF without character or codepoint label.')
        elif len(default_codepoint) > 1:
            logging.warning('Cannot store grapheme sequence as a BDF default glyph.')
        else:
            xlfd_props['DEFAULT_CHAR'] = default_codepoint[0]
        # unicode encoding
        xlfd_props['CHARSET_REGISTRY'] = '"ISO10646"'
        xlfd_props['CHARSET_ENCODING'] = '"1"'
    else:
        default_codepoint = default_glyph.codepoint
        if not len(default_codepoint):
            logging.debug('Cannot store default glyph in BDF without character or codepoint label.')
        else:
            xlfd_props['DEFAULT_CHAR'] = int(default_codepoint)
        # try preferred name
        encoding_name = _UNIX_ENCODINGS.get(font.encoding, font.encoding)
        registry, *encoding = encoding_name.split('-', 1)
        # encoding
        xlfd_props['CHARSET_REGISTRY'] = _quoted_string(registry.upper())
        if encoding:
            xlfd_props['CHARSET_ENCODING'] = _quoted_string(encoding[0].upper())
        else:
            xlfd_props['CHARSET_ENCODING'] = '"0"'
    # remove unset properties
    xlfd_props = {_k: _v for _k, _v in xlfd_props.items() if _v is not None}
    # keep unparsed properties
    unparsed = {
        _k.replace('-', '_').upper():
            _quoted_string(' '.join(_v.splitlines()))
        for _k, _v in font.get_properties().items()
        if not font.is_known_property(_k)
    }
    xlfd_props.update({
        _k.removeprefix('XLFD.'): _v
        for _k, _v in unparsed.items()
        if _k.startswith('XLFD.')
    })
    # keep unknown properties
    xlfd_props.update({
        _k.removeprefix(f'{CUSTOM_PROP}.'.upper()): _v
        for _k, _v in unparsed.items()
        if not _k.startswith('XLFD.')
    })
    return xlfd_props



"""
monobit.formats.mac.fond - Mac FOND font directory

(c) 2019--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import accumulate

from ..sfnt import MAC_ENCODING, mac_style_name, to_postscript_name
from ...binary import bytes_to_bits, align
from ...struct import bitfield, big_endian as be, sizeof
from ...glyph import Glyph


##############################################################################
# FOND resource
# https://developer.apple.com/library/archive/documentation/mac/Text/Text-269.html#MARKER-2-525

_FFLAGS = be.Struct(
    # bit 15: This bit is set to 1 if the font family describes fixed-width fonts, and is cleared
    #         to 0 if the font describes proportional fonts.
    fixed_width=bitfield('uint16', 1),
    # bit 14: This bit is set to 1 if the family fractional-width table is not used, and is cleared
    #         to 0 if the table is used.
    frac_width_unused=bitfield('uint16', 1),
    # bit 13: This bit is set to 1 if the font family should use integer extra width for stylistic
    #         variations. If not set, the font family should compute the fixed-point extra width
    #         from the family style-mapping table, but only if the FractEnable global variable
    #         has a value of TRUE.
    use_int_extra_width=bitfield('uint16', 1),
    # bit 12: This bit is set to 1 if the font family ignores the value of the FractEnable global
    #         variable when deciding whether to use fixed-point values for stylistic variations;
    #         the value of bit 13 is then the deciding factor. The value of the FractEnable global
    #         variable is set by the SetFractEnable procedure.
    ignore_global_fract_enable=bitfield('uint16', 1),
    # bits 2-11: These bits are reserved by Apple and should be cleared to 0.
    reserved_2_11=bitfield('uint16', 10),
    # bit 1: This bit is set to 1 if the resource contains a glyph-width table.
    has_width_table=bitfield('uint16', 1),
    # bit 0: This bit is reserved by Apple and should be cleared to 0.
    reserved_0=bitfield('uint16', 1),
)

# p1-110
# actually these are all signed??
_FOND_HEADER = be.Struct(
    # {flags for family}
    ffFlags=_FFLAGS,
    # {family ID number}
    ffFamID='uint16',
    # {ASCII code of first character}
    ffFirstChar='uint16',
    # {ASCII code of last character}
    ffLastChar='uint16',
    # {maximum ascent for 1-pt font}
    ffAscent='uint16',
    # {maximum descent for 1-pt font}
    ffDescent='uint16',
    # {maximum leading for 1-pt font}
    ffLeading='uint16',
    # {maximum glyph width for 1-pt font}
    ffWidMax='uint16',
    # {offset to family glyph-width table}
    ffWTabOff='uint32',
    # {offset to kerning table}
    ffKernOff='uint32',
    # {offset to style-mapping table}
    ffStylOff='uint32',
    # {style properties info}
    ffProperty=be.uint16 * 9,
    # {for international use}
    # reserved for internal use by script management software
    ffIntl=be.uint16 * 2,
    # {version number}
    ffVersion='uint16',
)

# font association table
# definitions I.M. p4-110
# signed??
_FA_HEADER = be.Struct(
    # number of entries - 1
    numAssoc='uint16',
)
# record - p4-111
_FA_ENTRY =  be.Struct(
    fontSize='uint16',
    fontStyle='uint16',
    fontID='uint16',
)

# offset table
# Fig 4-15, I. M.: Text p. 4-96
# will max_entry be -1 for an empty table?
_OFFS_HEADER = be.Struct(
    max_entry='int16',
)
# followed by array of uint32 offsets
_OFFS_ENTRY = be.Struct(
    offset='uint32',
)

# p. 4-99
# > Kerning distance. The kerning distance, in pixels, for the two glyphs
# > at a point size of 1. This is a 16-bit fixed point value, with the
# > integer part in the high-order 4 bits, and the fractional part in
# > the low-order 12 bits. The Font Manager measures the distance in pixels
# > and then multiplies it by the requested point size
_FIXED_TYPE = be.int16

# this leaves open the matter of negative numbers. there are two reasonable conventions
# 1) true value * 2**12 represented as a two's complement 16-bit integer
# 2) sign | 3-bit integer part | 12-bit fractional part
# which agree for positive numbers but are different for negatives.
# of course, both conventions are used:
# Palatino uses convention (2) while Helvetica uses (1).

# according to Adobe's 091.Mac_Fond.pdf (p.8):
# > If the value of this word is 2 or greater, then
# > negative numbers in the FOND resource are kept
# > as 2’s complement; font version 0 and 1 have 1’s
# > complement negative values.
# Palatino (1's complement) has ffVersion=0, Helvetica and Times (2's complement) have ffVersion=3
# mnany have ffVersion=1 but no kerning table

def fixed_to_float(fixed, twos_complement=True):
    # fixed is the input 2's complement 16-bit signed integer
    if not twos_complement:
        # convert back from 2's complement (interpreted by the C struct reader)
        fixed = 0x10000 + fixed
        # unset sign bit
        fixed = 0x7fff & fixed
        # make negative
        fixed = - fixed
    return fixed / 2**12

def _float_to_fixed(flt):
    # using convention (1)
    return int(flt * 2**12)


# bounding-box table
# Fig. 4.26
_BBX_HEADER = _OFFS_HEADER
_BBX_ENTRY = be.Struct(
    # _STYLE_MAP bitfield
    style='uint16',
    left=_FIXED_TYPE,
    bottom=_FIXED_TYPE,
    right=_FIXED_TYPE,
    top=_FIXED_TYPE,
)

# Family glyph width table
# definitions I.M. p.4-48, 4-98
_WIDTH_TABLE = be.Struct(
    # number of entries -1
    numWidths='int16',
)
_WIDTH_ENTRY = be.Struct(
    # style code
    widStyle='int16',
    # widths: ARRAY[0.n] of Fixed;
)

# Style-mapping table
# I.M.:Text p. 4-99
# > The font name suffix subtable and the glyph-encoding subtable that are part of the style-mapping
# > table immediately follow it in the resource data. The font name suffix subtable contains the
# > base font name and the suffixes that can be added to the font family’s name to produce a real
# > PostScript name (one that is recognized by the PostScript LaserWriter printer driver). The
# > style-mapping table uses the suffix table to build a font name for a PostScript printer. The
# > glyph-encoding table allows character codes to be mapped to PostScript glyph names.
_STYLE_TABLE = be.Struct(
    # bit field holding rendering hints - see I.M. p 4-101
    fontClass='int16',
    # offset from the start of this table to the glyph-encoding subtable component
    offset='int32',
    reserved='int32',
    # indexes into the font suffix name table that follows this table
    # "This is an array of 48 integer index values"
    # note C summary has 47 but Pascal summary has 0..47 inclusive
    indexes=be.int8 * 48,
)
# https://www6.uniovi.es/cscene/CS5/CS5-04.html
# > In Pascal, on the other hand, the first character of the string is the length of the
# > string, and the string is stored in the 255 characters that follow
# > On the Mac, there is a predefined type for Pascal strings, namely, Str255.
_STR255 = be.Struct(
    length='uint8',
    string=be.char * 255, #*length
)
_NAME_TABLE = be.Struct(
    stringCount='int16',
    #baseFontName=_STR255,
)

def string_from_pascal(data, offset):
    """Convert Mac-encoded Pascal string to Python string."""
    length = int(be.uint8.from_bytes(data, offset))
    string = data[offset+1:offset+1+length]
    string = string.decode('mac-roman', 'replace')
    return string, offset+1+length

def string_to_pascal(str):
    """Convert Python string to Mac-encoded Pascal string."""
    return bytes((len(str),)) + str.encode('mac-roman', 'replace')


# glyph encoding subtable
_ENC_TABLE = be.Struct(
    stringCount='int16',
)
_ENC_ENTRY = be.Struct(
    char='uint8',
    name=be.char * 255,
)

# Kerning table
_KERN_TABLE = be.Struct(
    # number of entries - 1
    numKerns='int16',
)
_KERN_ENTRY = be.Struct(
    kernStyle='uint16',
    kernLength='uint16',
)
_KERN_PAIR = be.Struct(
    kernFirst='uint8',
    kernSecond='uint8',
    # kerning value in 1pt fixed format
    kernWidth=_FIXED_TYPE,
)

def extract_fond(data, offset):
    """Read a MacOS FOND resource."""
    fond_header = _FOND_HEADER.from_bytes(data, offset)
    # Font Family Tables:
    # Font Association table (mandatory)
    fa_offset = offset + _FOND_HEADER.size
    fa_header = _FA_HEADER.from_bytes(data, fa_offset)
    fa_list = _FA_ENTRY.array(fa_header.numAssoc+1).from_bytes(data, fa_offset + _FA_HEADER.size)
    kerning_table = {}
    encoding_table = {}
    offs_header = None
    offs_list = None
    bbx_header = None
    bbx_list = None
    wtab = ()
    wtables = ()
    stab = ()
    names = ()
    ktab = None
    # check if any optional tables are expected
    # we don't have a field for bounding-box table offset
    if fond_header.ffWTabOff or fond_header.ffKernOff or fond_header.ffStylOff:
        # Offset table (optional, not used by us)
        # > Whenever any table, including the glyph-width, kerning, and
        # > style-mapping tables, is included in the resource data, an offset table is included.
        # > The offset table contains a long integer offset value for each table that follows it
        offs_offset = fa_offset + _FA_HEADER.size + _FA_ENTRY.size * (fa_header.numAssoc+1)
        offs_header = _OFFS_HEADER.from_bytes(data, offs_offset)
        # max_entry==-1 if the table is absent?
        offs_list = _OFFS_ENTRY.array(offs_header.max_entry+1).from_bytes(
            data, offs_offset + _OFFS_HEADER.size
        )
        # Bounding-box table (optional, not used by us)
        # no offset given in font record. should use the Offset Table to find it?
        # here we just assume it is the first table after the offset table
        # is BBX table effectively mandatory if more tables follow?
        bbx_offset = offs_offset + _OFFS_HEADER.size + _OFFS_ENTRY.size * (offs_header.max_entry+1)
        bbx_header = _BBX_HEADER.from_bytes(data, bbx_offset)
        bbx_list = _BBX_ENTRY.array(bbx_header.max_entry+1).from_bytes(
            data, bbx_offset + _BBX_HEADER.size
        )
        # Family glyph-width table (optional, not used by us)
        # use offset given in FOND header
        if fond_header.ffWTabOff:
            wtab_offset = offset + fond_header.ffWTabOff
            wtab = _WIDTH_TABLE.from_bytes(data, wtab_offset)
            # number of characters in each width table - one more than in NFNT.
            # poorly documented, but consistent with Classic Mac system resource
            # and FONDU/FontForge
            num_chars = fond_header.ffLastChar - fond_header.ffFirstChar + 3
            wtables = []
            wtab_offset += _WIDTH_TABLE.size
            # numWidths is the number of width tables, i.e. the number of styles.- 1
            for style in range(wtab.numWidths+1):
                wentry = _WIDTH_ENTRY.from_bytes(data, wtab_offset)
                widths = _FIXED_TYPE.array(num_chars).from_bytes(
                    data, wtab_offset + _WIDTH_ENTRY.size
                )
                wtab_offset += (
                    _WIDTH_ENTRY.size + _FIXED_TYPE.size * num_chars
                )
                wtables.append((wentry, widths))
        # Style-mapping table (optional)
        if fond_header.ffStylOff:
            stab_offset = offset + fond_header.ffStylOff
            stab = _STYLE_TABLE.from_bytes(data, stab_offset)
            # font name suffix subtable
            ntab_offset = stab_offset + _STYLE_TABLE.size
            ntab = _NAME_TABLE.from_bytes(data, ntab_offset)
            names = []
            offs = ntab_offset + _NAME_TABLE.size
            # count + 1 as we take the base font name as well?
            # but using that leads to incorrect encoding table
            for i in range(ntab.stringCount):
                string, offs = string_from_pascal(data, offs)
                names.append(string)
            # glyph-name encoding subtable
            etab_offset = offs
            etab = _ENC_TABLE.from_bytes(data, etab_offset)
            offs += _ENC_TABLE.size
            # encoding table - this is based on description in the docs
            # but does not appear to work correctly for built in mac Palatino font
            # which has a table that's just three nulls but stringCount=3
            for i in range(etab.stringCount):
                codepoint = be.uint8.from_bytes(data, offs)
                string, offs = string_from_pascal(data, offs+1)
                encoding_table[codepoint] = string
        # Kerning table (optional)
        if fond_header.ffKernOff:
            ktab_offset = offset + fond_header.ffKernOff
            ktab = _KERN_TABLE.from_bytes(data, ktab_offset)
            offs = ktab_offset + _KERN_TABLE.size
            kerning_table = {}
            for entry in range(ktab.numKerns+1):
                ke = _KERN_ENTRY.from_bytes(data, offs)
                # This is an integer value that specifies
                # the number of bytes in this kerning subtable
                pair_array = _KERN_PAIR.array(ke.kernLength)
                kerning_table[ke.kernStyle] = tuple(
                    pair_array.from_bytes(data, offs + _KERN_ENTRY.size)
                )
                offs += _KERN_ENTRY.size + pair_array.size
    return dict(
        fond_header=fond_header,
        fa_header=fa_header, fa_list=fa_list,
        offs_header=offs_header, offs_list=offs_list,
        bbx_header=bbx_header, bbx_list=bbx_list,
        width_header=wtab, width_list=wtables,
        style_header=stab, name_list=names,
        # do not include encoding table as I don't understand how it is structured
        #encoding_table=encoding_table,
        kerning_header=ktab, kerning_table=kerning_table,
    )


def convert_fond(
        name, fond_header, fa_list,
        kerning_table, encoding_table=None,
        **kwargs
    ):
    """Partially convert FOND properties to monobit properties."""
    # Inside Macintosh: Text 6-22
    # > Fonts with IDs below 16384 ($4000) are all Roman; starting with
    # > 16384 each non-Roman script system has a range of 512 ($200) font IDs available
    encoding = MAC_ENCODING.get(max(0, 1 + (fond_header.ffFamID - 16384) // 512))
    info = {
        # rsrc_id
        fa_entry.fontID: {
            # monobit properties
            'family': name,
            'style': mac_style_name(fa_entry.fontStyle),
            'point_size': fa_entry.fontSize,
            #'spacing': 'monospace' if fond_header.ffFlags.fixed_width else 'proportional',
            'encoding': encoding,
            # for use by NFNT parser
            'kerning-table': kerning_table.get(fa_entry.fontStyle, ()),
            'encoding-table': encoding_table,
            'twos-complement': fond_header.ffVersion >= 2,
        }
        for fa_entry in fa_list
    }
    # check if we're losing kerning tables
    styles = set(fa_entry.fontStyle for fa_entry in fa_list)
    dropped_styles = tuple(
        _style for _style in kerning_table
        if _style not in styles
    )
    for style in dropped_styles:
        logging.warning(
            'Kerning table not preserved '
            f'for style {style:#0x} ({mac_style_name(style)})'
        )
        logging.debug(kerning_table[style])
    return info


###############################################################################
# FOND writer

def create_fond(style_groups, family_id):
    """
    Convert monobit properties to FOND.
    Requires NFNT data structure for largest font.
    """
    # use last (largest) font in first group as representative sample
    largest_per_style = tuple(
        (_style_code, _font)
        for _style_code, (*_, _font) in style_groups
    )
    _, sample_font = largest_per_style[0]
    # list of (style_id, list of fonts)
    # following FONDU/FontForge: 0x9000 for fixed, 0x1000 for proportional
    ff_flags = _FFLAGS(
        fixed_width=sample_font.spacing in ('character-cell', 'monospace'),
        # bit 14: This bit is set to 1 if the family fractional-width table is not used, and is cleared
        #         to 0 if the table is used.
        frac_width_unused=0,
        # bit 13: This bit is set to 1 if the font family should use integer extra width for stylistic
        #         variations. If not set, the font family should compute the fixed-point extra width
        #         from the family style-mapping table, but only if the FractEnable global variable
        #         has a value of TRUE.
        use_int_extra_width=0,
        # bit 12: This bit is set to 1 if the font family ignores the value of the FractEnable global
        #         variable when deciding whether to use fixed-point values for stylistic variations;
        #         the value of bit 13 is then the deciding factor. The value of the FractEnable global
        #         variable is set by the SetFractEnable procedure.
        ignore_global_fract_enable=1,
        # bit 1: This bit is set to 1 if the resource contains a glyph-width table.
        # has_width_table=1,
    )
    fond_header = _FOND_HEADER(
        # {flags for family}
        ffFlags=ff_flags,
        # {family ID number}
        ffFamID=family_id,
        # {ASCII code of first character}
        ffFirstChar=int(min(sample_font.get_codepoints())),
        # {ASCII code of last character}
        ffLastChar=int(max(sample_font.get_codepoints())),
        # {maximum ascent for 1-pt font}
        ffAscent=_float_to_fixed(sample_font.ink_bounds.top / sample_font.bounding_box.y),
        # {maximum descent for 1-pt font}
        ffDescent=_float_to_fixed(sample_font.ink_bounds.bottom / sample_font.bounding_box.y),
        # {maximum leading for 1-pt font}
        ffLeading=_float_to_fixed(sample_font.line_height / sample_font.bounding_box.y - 1),
        # {maximum glyph width for 1-pt font}
        ffWidMax=_float_to_fixed(sample_font.max_width / sample_font.bounding_box.y),
        # {style properties info}
        # extra width for text styles (fixed point value) - currently left at all 0s
        #ffProperty=be.uint16 * 9,
        # {version number}
        # An integer value that specifies the version number of the font family resource, which indicates whether certain tables are available. This value is represented by the ffVersion field in the FamRec data type. Because this field has been used inconsistently in the system software, it is better to analyze the data in the resource itself instead of relying on the version number. The possible values are as follows:
        # Value	Meaning
        # $0000	Created by the Macintosh system software. The font family resource will not have the glyph-width tables and the fields will contain 0.
        # $0001	Original format as designed by the font developer. This font family record probably has the width tables and most of the fields are filled.
        # $0002	This record may contain the offset and bounding-box tables.
        # $0003	This record definitely contains the offset and bounding-box tables.
        # Adobe docs 0091.Mac_Fond.pdf:
        # > If the value of this word is 2 or greater, then negative numbers in
        # > the FOND resource are kept as 2’s complement;
        # > font version 0 and 1 have 1’s complement negative values.
        ffVersion=2,
    )
    fa_list = tuple(
        _FA_ENTRY(
            fontSize=_font.point_size,
            fontStyle=_style,
            # fontID=family_id+_i,
        )
        for _style, _group in style_groups
        for _font in _group
    )
    # following FONDU, just increment from the family id.
    # not sure if/how this avoids ID collisions
    # NOTE this is calculated separately for NFNT's rsrc id
    for i, entry in enumerate(fa_list):
        entry.fontID = family_id + i
    fa_list = (_FA_ENTRY * len(fa_list))(*fa_list)
    fa_header = _FA_HEADER(numAssoc=len(fa_list)-1)
    # get contiguous vector of glyphs, add missing glyph and one extra with 1-em width
    # the latter is described as 'Unused Word' in Adobe docs
    style_groups = tuple(
        (
            _style_id,
            tuple(
                _font.modify(glyphs=(
                    [
                        _font.get_glyph(_cp, missing='empty')
                        for _cp in range(fond_header.ffFirstChar, fond_header.ffLastChar+1)
                    ]
                    + [_font.glyphs[-1], Glyph(scalable_width=_font.bounding_box.y)]
                ))
                for _font in _group
            )
        )
        for _style_id, _group in style_groups
    )
    # optional tables
    bbx_bytes = _create_bbx_table(style_groups)
    wtab_bytes = _create_width_table(style_groups)
    stab_bytes = _create_style_table(style_groups)
    ktab_bytes = _create_kerning_table(style_groups)
    optional_tables = (bbx_bytes, wtab_bytes, stab_bytes, ktab_bytes)
    # # Offset table (mandatory if optional tables are present)
    # one entry for each table, with its offsets
    # offset from the start of the offset table
    num_tables = len(optional_tables)
    offset_table_size = _OFFS_HEADER.size + num_tables*_OFFS_ENTRY.size
    # > the number of bytes from the start of the offset table to the start of the table.
    table_offsets = tuple(accumulate(
        (len(_t) for _t in optional_tables),
        initial=offset_table_size
    ))
    bbx_offset, wtab_offset, stab_offset, ktab_offset, _ = table_offsets
    table_offsets = (_OFFS_ENTRY * num_tables)(*(
        _OFFS_ENTRY(offset=_ofs) for _ofs in table_offsets[:-1]
    ))
    offs_header = _OFFS_HEADER(max_entry=num_tables-1)
    offs_bytes = bytes(offs_header) + bytes(table_offsets)
    # offset fields in header
    # > The offset to the family glyph-width table from the beginning of
    # > the font family resource to the beginning of the table, in bytes.
    header_size = _FOND_HEADER.size + _FA_HEADER.size + sizeof(fa_list)
    fond_header.ffWTabOff = wtab_offset + header_size
    fond_header.ffStylOff = stab_offset + header_size
    fond_header.ffKernOff = ktab_offset + header_size
    return (
        bytes(fond_header) + bytes(fa_header) + bytes(fa_list)
        + offs_bytes + b''.join(optional_tables)
    )


def _create_bbx_table(style_groups):
    """Bounding-box table (optional, but no clear way to indicate if it's present)"""
    bbx_header = _BBX_HEADER(max_entry=len(style_groups)-1,)
    # we need one bounding box entry per style
    # metrics are scalable given as fixed-point fraction for a 1pt font
    # N.B. FONDU seems to just put integer pixel values here
    bbx_list = tuple(
        _BBX_ENTRY(
            style=_style_code,
            left=_float_to_fixed(_font.ink_bounds.left / _font.point_size),
            bottom=_float_to_fixed(_font.ink_bounds.bottom / _font.point_size),
            right=_float_to_fixed(_font.ink_bounds.right / _font.point_size),
            top=_float_to_fixed(_font.ink_bounds.top / _font.point_size),
        )
        # use last (largest) font for each style
        for _style_code, (*_, _font) in style_groups
    )
    bbx_entries = (_BBX_ENTRY * len(bbx_list))(*bbx_list)
    bbx_bytes = bytes(bbx_header) + bytes(bbx_entries)
    return bbx_bytes


def _create_width_table(style_groups):
    """Family glyph-width table (optional)"""
    wtab = _WIDTH_TABLE(numWidths=len(style_groups)-1)
    wtables = tuple(
        (
            _WIDTH_ENTRY(widStyle=_style_id),
            _FIXED_TYPE.array(len(_group[-1].glyphs))(*(
                _float_to_fixed(_g.scalable_width / _group[-1].bounding_box.y)
                for _g in _group[-1].glyphs
            ))
        )
        for _style_id, _group in style_groups
    )
    wtab_bytes = bytes(wtab) + b''.join(
        bytes(_t) for _pair in wtables for _t in _pair
    )
    return wtab_bytes


def _create_style_table(style_groups):
    """Style-mapping table (optional). This maps printer fonts to screen fonts."""
    # # font name suffix subtable
    suffixes = ('-',) + tuple(sorted(set(
        to_postscript_name(_suffix)
        for _, _group in style_groups
        for _suffix in _group[-1].subfamily.split()
    )))
    used_suffixes_per_style = tuple(
        tuple(
            suffixes.index(_suffix)
            for _suffix in _group[-1].subfamily.split()
        )
        for _, _group in style_groups
    )
    used_style_ids = tuple(_id for _id, _ in style_groups)
    suffix_subtable = (
        # this is the encoding of the postscript name for a given style
        # [number of name parts] followed by pointer to name parts in this table,
        # counting from 1, with element 1 (the family name) implicit
        chr(len(_used_suffixes)+1) + ''.join(chr(2+_x) for _x in _used_suffixes)
        for _used_suffixes in used_suffixes_per_style
    )
    stringtable = (
        to_postscript_name(style_groups[-1][1][-1].family),
        *suffix_subtable,
        *suffixes
    )
    # convert to P-strings
    stringtable = tuple(string_to_pascal(_str) for _str in stringtable)
    ntab = _NAME_TABLE(stringCount=len(stringtable))
    # indexes in the style name table count from 1, for some reason (Adobe docs):
    # > Because the Suffix List is part of the Style Name Table, these indexes count
    # > from the beginning of the Style Name Table. In particular, note that the
    # > number of strings counts as an entry and has index value 0.
    style_ids = (
        _style_mapping_to_style_id(_mapping_value)
        for _mapping_value in range(48)
    )
    indexes = []
    for id in style_ids:
        try:
            indexes.append(1 + used_style_ids.index(id))
        except ValueError:
            # we repeat the last index for unused entries. Adobe docs seem to suggest
            # these could be 0 but that crashes FontForge
            # FontForge treats repeated values as a not-in-use sentinel.
            indexes.append(indexes[-1] if indexes else 0)
    # # glyph-name encoding subtable
    # generate empty encoding table - I don't know how to construct correctly
    # and FontForge code comments suggest that FontManager rejects fonts that have this table.
    etab = _ENC_TABLE(stringCount=0)
    # bit field holding rendering hints - see I.M. p 4-101
    # > 0 This bit is set to 1 if the font name needs coordinating.
    # following FONDU, allow simulated style if not explicitly defined
    font_class = 1<<0
    if any('outline' in _f.decoration for _, _group in style_groups for _f in _group):
        font_class &= 1<<3
    if any(_f.weight == 'bold' for _, _group in style_groups for _f in _group):
        font_class &= 1<<4
    if any(_f.slant == 'italic' for _, _group in style_groups for _f in _group):
        font_class &= 1<<6
    if any(_f.setwidth == 'condensed' for _, _group in style_groups for _f in _group):
        font_class &= 1<<7
    if any(_f.setwidth == 'expanded' for _, _group in style_groups for _f in _group):
        font_class &= 1<<8
    # putting together the style-mapping table
    stab = _STYLE_TABLE(
        fontClass=font_class,
        # offset from the start of this table to the glyph-encoding subtable component
        offset=_STYLE_TABLE.size + sizeof(ntab) + sum(len(_s) for _s in stringtable),
        # indexes into the font suffix name table that follows this table
        # "This is an array of 48 integer index values"
        # note C summary has 47 but Pascal summary has 0..47 inclusive
        indexes=(be.int8 * 48)(*indexes),
    )
    stab_bytes = bytes(stab) + bytes(ntab) + b''.join(stringtable) + bytes(etab)
    return stab_bytes


def _style_mapping_to_style_id(mapping_value):
    """Convert Mac style id to Style Mapping used for postscript table."""
    # mapping removes bit 2, underline
    # see 0091.Mac_Fond.pdf p13 table 2
    return ((mapping_value & 0b111100) << 1) + (mapping_value & 0b11)



def _create_kerning_table(style_groups):
    """Kerning table (optional).
    """
    num_styles = len(style_groups)
    ktab = _KERN_TABLE(numKerns=num_styles-1)
    ktab_bytes = bytearray(bytes(ktab))
    for style_id, group in style_groups:
        # NOTE that the format assumes one kerning table per style,
        # which is linearly scaled by bounding_box.y
        sample_font = group[-1]
        kerning_pairs = [
            _KERN_PAIR(
                kernFirst=int(_g.codepoint),
                kernSecond=int(sample_font.get_glyph(_label).codepoint),
                # kerning value in 1pt fixed format
                kernWidth=_float_to_fixed(_value / sample_font.bounding_box.y),
            )
            for _g in sample_font.glyphs
            for _label, _value in _g.right_kerning.items()
        ] + [
            _KERN_PAIR(
                kernFirst=int(sample_font.get_glyph(_label).codepoint),
                kernSecond=int(_g.codepoint),
                # kerning value in 1pt fixed format
                kernWidth=_float_to_fixed(_value / sample_font.bounding_box.y),
            )
            for _g in sample_font.glyphs
            for _label, _value in _g.left_kerning.items()
        ]
        # number of entries - 1
        ke = _KERN_ENTRY(
            kernStyle=style_id,
            kernLength=len(kerning_pairs),
        )
        kpairs = (_KERN_PAIR * len(kerning_pairs))(*kerning_pairs)
        ktab_bytes += bytes(ke) + bytes(kpairs)
    return bytes(ktab_bytes)



"""
monobit.formats.mac - Mac OS fonts

(c) 2019--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from ...storage import loaders, savers

from .dfont import parse_resource_fork, save_dfont
from .nfnt import extract_nfnt, convert_nfnt, create_nfnt
from .lisa import _load_lisa
from .iigs import _load_iigs, _save_iigs


@loaders.register(
    name='mac',
    # the magic is optional - a 'maybe magic'
    magic=(b'\0\0\1\0\0',),
    patterns=('*.dfont', '*.suit', '*.rsrc',),
)
def load_mac_dfont(instream):
    """Load font from MacOS resource fork or data-fork resource."""
    data = instream.read()
    return parse_resource_fork(data)


@savers.register(linked=load_mac_dfont)
def save_mac_dfont(fonts, outstream, resource_type:str='NFNT', family_id:int=None):
    """Save font to MacOS resource fork or data-fork resource.

    resource_type: type of resource to store font in. One of `sfnt`, `NFNT`.
    """
    save_dfont(fonts, outstream, resource_type)


@loaders.register(
    name='nfnt',
    # \x90\0 is not a formal signature, but the most common set of FONT_TYPE flags
    # the \x80 sigs are LISA compressed NFNTs
    magic=(b'\x90\0', b'\xb0\0', b'\x90\x80', b'\xb0\x80'),
    patterns=('*.f',),
)
def load_nfnt(instream, offset:int=0):
    """
    Load font from a bare FONT/NFNT resource.

    offset: starting offset in bytes of the NFNT record in the file (default 0)
    """
    instream.seek(offset)
    data = instream.read()
    fontdata = extract_nfnt(data, 0)
    return convert_nfnt({}, **fontdata)


@loaders.register(name='lisa')
def load_lisa(instream):
    """Load a LISA font library."""
    return _load_lisa(instream)


@loaders.register(
    name='iigs',
    patterns=('*.fon',),
)
def load_iigs(instream):
    """Load a IIgs font."""
    return _load_iigs(instream)


@savers.register(linked=load_iigs)
def save_iigs(fonts, outstream, version:int=None):
    """
    Write font to a IIgs font file.

    version: IIgs font format version (0x101, 0x105). Default: 0x101 unless needed for bitmap size.
    """
    if len(fonts) > 1:
        logging.warning('IIgs font file can only store one font.')
    font = fonts[0]
    _save_iigs(outstream, font, version=version)


@savers.register(linked=load_nfnt)
def save_nfnt(
        fonts, outstream,
        create_width_table:bool=True,
        create_height_table:bool=False,
    ):
    """
    Write font to a bare FONT/NFNT resource.

    create_width_table: include a fractional glyph-width table in the resource (default: True)
    create_height_table: include an image-height table in the resource (default: False)
    """
    if len(fonts) > 1:
        logging.warning('NFNT resource can only store one font.')
    font = fonts[0]
    data, _, _ = create_nfnt(
        font, endian='big', ndescent_is_high=True,
        create_width_table=create_width_table,
        create_height_table=create_height_table,
    )
    outstream.write(data)



"""
monobit.formats.xlfd.pcf - X11 portable compiled format

(c) 2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import accumulate

from ...struct import big_endian as be, little_endian as le
from ...storage import loaders, savers
from ...magic import FileFormatError
from ...properties import Props
from ...font import Font
from ...glyph import Glyph
from ...raster import Raster
from ...labels import Tag, Codepoint
from ...binary import align, ceildiv

from .bdf import swidth_to_pixel, pixel_to_swidth
from .xlfd import (
    parse_xlfd_properties, create_xlfd_properties, create_xlfd_name,
    from_quoted_string
)

MAGIC = b'\1fcp'

@loaders.register(
    name='pcf',
    magic=(MAGIC,),
    patterns=('*.pcf',),
)
def load_pcf(instream):
    """Load font from X11 Portable Compiled Format (PCF)."""
    pcf_data = _read_pcf(instream)
    glyphs = _convert_glyphs(pcf_data)
    props = _convert_props(pcf_data)
    font = Font(glyphs, **props)
    return font.label()

@savers.register(linked=load_pcf)
def save_pcf(
        fonts, outstream, *,
        byte_order:str='big', ink_bounds:bool=True, scan_unit:int=1,
        padding_bytes:int=1, bit_order:str='big',
    ):
    """
    Save font to X11 Portable Compiled Format (PCF).

    ink_bounds: include optional ink-bounds metrics (default: True)
    byte_order: 'big'-endian (default) or 'little'-endian
    bit_order: 'big'-endian (default) if left pixel is most significant bit, or 'little'-endian if left is lsb
    scan_unit: number of bytes per unit in bitmap (1, 2, 4 or 8; default is 1)
    padding_bytes: make raster row a multiple of this number of bytes (1, 2, 4 or 8; default is 1)
    """
    font, *more = fonts
    if more:
        raise FileFormatError('Can only save one font to BDF file.')
    # can only do big-endian for now
    _write_pcf(
        outstream, font, endian=byte_order, create_ink_bounds=ink_bounds,
        scan_unit=scan_unit, padding_bytes=padding_bytes, bit_order=bit_order,
    )
    return font

##############################################################################
# https://fontforge.org/docs/techref/pcf-format.html

_HEADER = le.Struct(
    # /* always "\1fcp" */
    header='4s',
    table_count='int32',
)

# fontforge recap has these as apparent signed ints,
# but X sources say CARD32 which is an unsigned int
_TOC_ENTRY = le.Struct(
    # /* See below, indicates which table */
    type='uint32',
    # /* See below, indicates how the data are formatted in the table */
    format='uint32',
    # /* In bytes */
    size='uint32',
    # /* from start of file */
    offset='uint32',
)

# format field
#define PCF_DEFAULT_FORMAT       0x00000000
PCF_DEFAULT_FORMAT = 0x00000000
#define PCF_INKBOUNDS           0x00000200
#define PCF_ACCEL_W_INKBOUNDS   0x00000100
PCF_ACCEL_W_INKBOUNDS = 0x00000100
#define PCF_COMPRESSED_METRICS  0x00000100
PCF_COMPRESSED_METRICS = 0x00000100

# format field modifiers
#define PCF_GLYPH_PAD_MASK       (3<<0)            /* See the bitmap table for explanation */
PCF_GLYPH_PAD_MASK = (3<<0)
#define PCF_BYTE_MASK           (1<<2)            /* If set then Most Sig Byte First */
PCF_BYTE_MASK = (1<<2)
#define PCF_BIT_MASK            (1<<3)            /* If set then Most Sig Bit First */
PCF_BIT_MASK = (1<<3)
#define PCF_SCAN_UNIT_MASK      (3<<4)            /* See the bitmap table for explanation */
PCF_SCAN_UNIT_MASK = (3<<4)

# type field
#define PCF_PROPERTIES               (1<<0)
PCF_PROPERTIES = (1<<0)
#define PCF_ACCELERATORS            (1<<1)
PCF_ACCELERATORS = (1<<1)
#define PCF_METRICS                 (1<<2)
PCF_METRICS = (1<<2)
#define PCF_BITMAPS                 (1<<3)
PCF_BITMAPS = (1<<3)
#define PCF_INK_METRICS             (1<<4)
PCF_INK_METRICS = (1<<4)
#define PCF_BDF_ENCODINGS           (1<<5)
PCF_BDF_ENCODINGS = (1<<5)
#define PCF_SWIDTHS                 (1<<6)
PCF_SWIDTHS = (1<<6)
#define PCF_GLYPH_NAMES             (1<<7)
PCF_GLYPH_NAMES = (1<<7)
#define PCF_BDF_ACCELERATORS        (1<<8)
PCF_BDF_ACCELERATORS = (1<<8)


def _read_pcf(instream):
    """Read font from X11 PCF font file."""
    header = _HEADER.read_from(instream)
    toc = (_TOC_ENTRY * header.table_count).read_from(instream)
    props = Props()
    for entry in toc:
        instream.seek(entry.offset)
        if entry.type == PCF_PROPERTIES:
            props.xlfd_props = _read_properties_table(instream)
        elif entry.type == PCF_ACCELERATORS:
            # mandatory if BDF_ACCELERATORS not defined
            props.acc_props = _read_acc_table(instream)
        elif entry.type == PCF_BDF_ACCELERATORS:
            # optional
            props.bdf_acc_props = _read_acc_table(instream)
        elif entry.type == PCF_METRICS:
            # mandatory
            props.metrics = _read_metrics(instream)
        elif entry.type == PCF_INK_METRICS:
            # optional
            props.ink_metrics = _read_metrics(instream)
        elif entry.type == PCF_BITMAPS:
            # mandatory
            props.bitmap_format, props.bitmaps = _read_bitmaps(instream)
        elif entry.type == PCF_BDF_ENCODINGS:
            # mandatory, but could be empty
            props.encodings, props.default_char = _read_encoding(instream)
        elif entry.type == PCF_SWIDTHS:
            # optional - does not exist in X11 R6.4 sources
            props.swidths = _read_swidths(instream)
        elif entry.type == PCF_GLYPH_NAMES:
            # optional - does not exist in X11 R6.4 sources
            props.glyph_names = _read_glyph_names(instream)
    return props


def _read_format(instream):
    """Read the format record at start of tables."""
    format = int(le.uint32.read_from(instream))
    if format & PCF_BYTE_MASK:
        base = be
    else:
        base = le
    return format, base


# Properties table

# can be be or le
_PROPS = dict(
    name_offset='int32',
    isStringProp='int8',
    value='int32',
)

def _read_properties_table(instream):
    """Read the Properties table."""
    format, base = _read_format(instream)
    nprops = base.uint32.read_from(instream)
    props = (base.Struct(**_PROPS) * nprops).read_from(instream)
    #  pad to next int32 boundary
    padding = instream.read(0 if nprops&3 == 0 else 4-(nprops&3))
    string_size = base.int32.read_from(instream)
    strings = instream.read(string_size)
    xlfd_props = {}
    for prop in props:
        name, _, _ = strings[prop.name_offset:].partition(b'\0')
        name = name.decode('latin-1', 'ignore')
        if prop.isStringProp:
            value, _, _ = strings[int(prop.value):].partition(b'\0')
            value = value.decode('latin-1', 'ignore')
        else:
            value = int(prop.value)
        xlfd_props[name] = value
    return xlfd_props


# Glyph metrics and ink-metrics

# There are two different metrics tables, PCF_METRICS and PCF_INK_METRICS, the
# former contains the size of the stored bitmaps, while the latter contains the
# minimum bounding box. The two may contain the same data, but many CJK fonts
# pad the bitmaps so all bitmaps are the same size.

# from https://tronche.com/gui/x/xlib/graphics/font-metrics/#XCharStruct
# typedef struct {
# 	short lbearing;			/* origin to left edge of raster */
# 	short rbearing;			/* origin to right edge of raster */
# 	short width;			/* advance to next char's origin */
# 	short ascent;			/* baseline to top edge of raster */
# 	short descent;			/* baseline to bottom edge of raster */
# 	unsigned short attributes;	/* per char flags (not predefined) */
# } XCharStruct;

# X docs suggest a right-to-left character would have negative character_width (advance)
# it is unclear about what this means for bearings, as the raster width is given to be rbearing-lbearing
# in any case, in practice RTL chars appear to have positive metrics
# see e.g. cuarabic12.pcf, 6x13-ISO8859-8.pcf
# so we assume character width is always positive, and advance direction is
# decided by the renderer (as we do it)

_UNCOMPRESSED_METRICS = dict(
    left_side_bearing='int16',
    right_side_bearing='int16',
    character_width='int16',
    character_ascent='int16',
    character_descent='int16',
    character_attributes='uint16',
)

# The (compressed) bytes are unsigned bytes which are offset by 0x80
# (so the actual value will be (getc(pcf_file)-0x80). :
_COMPRESSED_METRICS = dict(
    left_side_bearing='uint8',
    right_side_bearing='uint8',
    character_width='uint8',
    character_ascent='uint8',
    character_descent='uint8',
)


# Accelerator table

_ACC_TABLE = dict(
    noOverlap='uint8',
    constantMetrics='uint8',
    terminalFont='uint8',
    constantWidth='uint8',
    inkInside='uint8',
    inkMetrics='uint8',
    drawDirection='uint8',
    padding='uint8',
    fontAscent='int32',
    fontDescent='int32',
    maxOverlap='int32',
    # minimum and maximum value for each metric
    #minbounds=_UNCOMPRESSED_METRICS,
    #maxbounds=_UNCOMPRESSED_METRICS,
    # if format PCF_ACCEL_W_INKBOUNDS:
    # maximum and maximum value for each ink metric
    ##ink_minbounds=_UNCOMPRESSED_METRICS,
    ##ink_maxbounds=_UNCOMPRESSED_METRICS,
)


def _read_acc_table(instream):
    """Read the Accelerator or BDF Accelerator table."""
    format, base = _read_format(instream)
    acc_table = base.Struct(**_ACC_TABLE).read_from(instream)
    acc_table = Props(**vars(acc_table))
    uncompressed_metrics = base.Struct(**_UNCOMPRESSED_METRICS)
    acc_table |= Props(
        minbounds=uncompressed_metrics.read_from(instream),
        maxbounds=uncompressed_metrics.read_from(instream),
    )
    if format & PCF_ACCEL_W_INKBOUNDS:
        acc_table |= Props(
            ink_minbounds=uncompressed_metrics.read_from(instream),
            ink_maxbounds=uncompressed_metrics.read_from(instream),
        )
    return acc_table


def _read_metrics(instream):
    """Read the Metrics or Ink-Metrics table."""
    format, base = _read_format(instream)
    if format & PCF_COMPRESSED_METRICS:
        compressed_metrics = base.Struct(**_COMPRESSED_METRICS)
        # documented as signed int, but unsigned it makes more sense
        # also this is used as uint by bdftopcf for e.g. unifont
        count = base.uint16.read_from(instream)
        metrics = (compressed_metrics * count).read_from(instream)
        # adjust unsigned bytes by 0x80 offset
        metrics = tuple(
            Props(**{_k: _v-0x80 for _k, _v in vars(_m).items()})
            for _m in metrics
        )
    else:
        uncompressed_metrics = base.Struct(**_UNCOMPRESSED_METRICS)
        count = base.uint32.read_from(instream)
        metrics = (uncompressed_metrics * count).read_from(instream)
    return metrics


def _read_bitmaps(instream):
    """Read the Bitmaps table."""
    format, base = _read_format(instream)
    glyph_count = base.int32.read_from(instream)
    offsets = (base.int32 * glyph_count).read_from(instream)
    bitmap_sizes = (base.int32 * 4).read_from(instream)
    bitmap_size = bitmap_sizes[format & 3]
    bitmap_data = instream.read(bitmap_size)
    offsets = tuple(offsets) + (None,)
    return format, tuple(
        bitmap_data[_offs:_next]
        for _offs, _next in zip(offsets, offsets[1:])
    )

# from https://tronche.com/gui/x/xlib/graphics/font-metrics/#XFontStruct
# typedef struct {
# 	XExtData *ext_data;		/* hook for extension to hang data */
# 	Font fid;			/* Font id for this font */
# 	unsigned direction;		/* hint about the direction font is painted */
# 	unsigned min_char_or_byte2;	/* first character */
# 	unsigned max_char_or_byte2;	/* last character */
# 	unsigned min_byte1;		/* first row that exists */
# 	unsigned max_byte1;		/* last row that exists */
# 	Bool all_chars_exist;		/* flag if all characters have nonzero size */
# 	unsigned default_char;		/* char to print for undefined character */
# 	int n_properties;		/* how many properties there are */
# 	XFontProp *properties;		/* pointer to array of additional properties */
# 	XCharStruct min_bounds;		/* minimum bounds over all existing char */
# 	XCharStruct max_bounds;		/* maximum bounds over all existing char */
# 	XCharStruct *per_char;		/* first_char to last_char information */
# 	int ascent;			/* logical extent above baseline for spacing */
# 	int descent;			/* logical decent below baseline for spacing */
# } XFontStruct;

# FontForge docs suggest the encoding table has signed integers
# but the XFontStruct has them unsigned
# they also make more sense unsigned, especially default_char which may be two-byte.
_ENCODING_TABLE = dict(
    min_char_or_byte2='uint16',
    max_char_or_byte2='uint16',
    min_byte1='uint16',
    max_byte1='uint16',
    default_char='uint16',
)

def _generate_codepoints(enc):
    """Generate all code points based on encoding table."""
    if not enc.min_byte1 and not enc.max_byte1:
        codepoints = (
            bytes((_cp,))
            for _cp in range(enc.min_char_or_byte2, enc.max_char_or_byte2+1)
        )
    else:
        codepoints = (
            bytes((_hi, _lo))
            for _hi in range(enc.min_char_or_byte2, enc.max_char_or_byte2+1)
            for _lo in range(enc.min_byte1, enc.max_byte1+1)
        )
    return codepoints

def _read_encoding(instream):
    format, base = _read_format(instream)
    enc = base.Struct(**_ENCODING_TABLE).read_from(instream)
    count = (enc.max_char_or_byte2-enc.min_char_or_byte2+1)*(enc.max_byte1-enc.min_byte1+1)
    # generate code points
    codepoints = _generate_codepoints(enc)
    glyph_indices = (base.int16 * count).read_from(instream)
    encoding_dict = {
        _cp: _idx for _cp, _idx in zip(codepoints, glyph_indices)
        # -1 means 'not used'
        if _idx >= 0
    }
    return encoding_dict, enc.default_char


def _read_swidths(instream):
    """Read the Scalable Widths table."""
    format, base = _read_format(instream)
    glyph_count = base.uint32.read_from(instream)
    swidths = (base.int32 * glyph_count).read_from(instream)
    return swidths


def _read_glyph_names(instream):
    """Read the Glyph Names table."""
    format, base = _read_format(instream)
    glyph_count = base.int32.read_from(instream)
    offsets = (base.int32 * glyph_count).read_from(instream)
    string_size = base.int32.read_from(instream)
    strings = instream.read(string_size)
    glyph_names = []
    for ofs in offsets:
        name, _, _ = strings[ofs:].partition(b'\0')
        name = name.decode('latin-1', 'ignore')
        glyph_names.append(name)
    return glyph_names


###############################################################################
# converter

def _convert_glyphs(pcf_data):
    """Convert glyphs from X11 PCF data to monobit."""
    # label sets
    n_glyphs = len(pcf_data.metrics)
    try:
        # if the table exists, the count should be the same as metrics count
        labelsets = [[Tag(_name)] for _name in pcf_data.glyph_names]
    except AttributeError:
        labelsets = [[] for _ in pcf_data.metrics]
    try:
        for _cp, _idx in pcf_data.encodings.items():
            labelsets[_idx].append(Codepoint(_cp))
    except AttributeError:
        pass
    # scalable width reference values
    if hasattr(pcf_data, 'swidths'):
        dpi_y = pcf_data.xlfd_props.get('RESOLUTION_Y', 72)
        try:
            point_size = pcf_data.xlfd_props['POINT_SIZE'] / 10
        except KeyError:
            try:
                point_size = pcf_data.xlfd_props['PIXEL_SIZE'] * dpi_y / 72 / 10
            except KeyError:
                logging.warning('No point-size information - dropping scalable width table')
                # can't calculate swidths
                pcf_data.swidths = None
    if hasattr(pcf_data, 'swidths') and pcf_data.swidths:
        pcf_data.swidths = tuple(
            swidth_to_pixel(_swidth, point_size, dpi_y)
            for _swidth in pcf_data.swidths
        )
    else:
        pcf_data.swidths = (None,) * n_glyphs
    # /* how each row in each glyph's bitmap is padded (format&3) */
    # /*  0=>bytes, 1=>shorts, 2=>ints */
    glyph_pad_length = pcf_data.bitmap_format & PCF_GLYPH_PAD_MASK
    byte_big = pcf_data.bitmap_format & PCF_BYTE_MASK
    bit_big = pcf_data.bitmap_format & PCF_BIT_MASK
    # /* what the bits are stored in (bytes, shorts, ints) (format>>4)&3 */
    # /*  0=>bytes, 1=>shorts, 2=>ints */
    scan_unit = (pcf_data.bitmap_format & PCF_SCAN_UNIT_MASK) >> 4
    # metrics are as defined in XCharStruct (above)
    # https://tronche.com/gui/x/xlib/graphics/font-metrics/
    # ascent and descent determine the character height, which we instead infer
    # from bitmap and stride.
    glyphs = tuple(
        Glyph.from_bytes(
            _gb,
            width=_met.right_side_bearing-_met.left_side_bearing,
            stride=align(_met.right_side_bearing-_met.left_side_bearing, glyph_pad_length+3),
            height=_met.character_ascent+_met.character_descent,
            bit_order='big' if bit_big else 'little',
            byte_swap=0 if (bool(byte_big) == bool(bit_big)) else 2**scan_unit,
            left_bearing=_met.left_side_bearing,
            right_bearing=_met.character_width-_met.right_side_bearing,
            shift_up=-_met.character_descent,
            scalable_width=_swidth if _swidth != _met.character_width else None,
            labels=_labs,
        )
        for _gb, _met, _labs, _swidth in zip(
            pcf_data.bitmaps, pcf_data.metrics, labelsets, pcf_data.swidths
        )
    )
    return glyphs


def _convert_props(pcf_data):
    """Convert properties for PCF to monobit."""
    xlfd_name = pcf_data.xlfd_props.pop('FONT', '')
    pcf_data.xlfd_props = {_k: str(_v) for _k, _v in pcf_data.xlfd_props.items()}
    props = parse_xlfd_properties(pcf_data.xlfd_props, xlfd_name)
    props.update(dict(
        default_char=Codepoint(pcf_data.default_char),
    ))
    # ascent and descent - these are stored in accelerator table rather than XLFD props
    if hasattr(pcf_data, 'bdf_acc_props'):
        props.update(dict(
            ascent=pcf_data.bdf_acc_props.fontAscent,
            descent=pcf_data.bdf_acc_props.fontDescent,
        ))
    elif hasattr(pcf_data, 'acc_props'):
        props.update(dict(
            ascent=pcf_data.acc_props.fontAscent,
            descent=pcf_data.acc_props.fontDescent,
        ))
    return props


###############################################################################
# pcf writer

def _write_pcf(
        outstream, font,
        endian, create_ink_bounds, scan_unit, padding_bytes, bit_order,
    ):
    """Write font to X11 PCF font file."""
    if endian[:1].lower() == 'b':
        base = be
    else:
        base = le
    # full format needs to be repeated in every table (at least the low byte)
    # or pcf2bdf won't recognise the bitmaps correctly
    format = PCF_DEFAULT_FORMAT
    if bit_order[:1].lower() == 'b':
        format |= PCF_BIT_MASK
    if base == be:
        format |= PCF_BYTE_MASK
    # /* how each row in each glyph's bitmap is padded (format&3) */
    # /*  0=>bytes, 1=>shorts, 2=>ints */
    # do a int(log2())
    format |= (padding_bytes.bit_length()-1) & PCF_GLYPH_PAD_MASK
    # /* what the bits are stored in (bytes, shorts, ints) (format>>4)&3 */
    # /*  0=>bytes, 1=>shorts, 2=>ints */
    format |= ((scan_unit.bit_length()-1) << 4) & PCF_SCAN_UNIT_MASK
    # tables MUST be in this order or pcf2bdf will reject the file
    tables = (
        (PCF_PROPERTIES, *_create_properties_table(font, format, base)),
        (PCF_ACCELERATORS, *_create_acc_table(font, format, base, create_ink_bounds)),
        (PCF_METRICS, *_create_metrics_table(font, format, base, _create_glyph_metrics)),
        # pcf2bdf doesn't read ink metrics, assume they go here?
        (PCF_INK_METRICS, *_create_metrics_table(font, format, base, _create_ink_metrics)),
        (PCF_BITMAPS, *_create_bitmaps(
            font, format, base,
            scan_unit_bytes=scan_unit, padding_bytes=padding_bytes,
            bit_order=bit_order,
        )),
        (PCF_BDF_ENCODINGS, *_create_encoding(font, format, base)),
        (PCF_SWIDTHS, *_create_swidths(font, format, base)),
        (PCF_GLYPH_NAMES, *_create_glyph_names(font, format, base)),
        # we're using the same table for accelerators and BDF accelerators
        # intended difference is unclear
        (PCF_BDF_ACCELERATORS, *_create_acc_table(font, format, base, create_ink_bounds)),
    )
    # calculate offsets, construct ToC
    offset = _HEADER.size + _TOC_ENTRY.size * len(tables)
    toc = []
    for type, table, format in tables:
        toc.append(_TOC_ENTRY(
            type=type,
            format=format,
            size=len(table),
            offset=offset,
        ))
        # align on 2**5==32-bit boundaries
        offset += align(len(table), 5)
    # write out file
    header = _HEADER(header=MAGIC, table_count=len(tables))
    outstream.write(bytes(header))
    for entry in toc:
        outstream.write(bytes(entry))
    for (_, table, _), next in zip(tables, toc[1:] + [None]):
        outstream.write(table)
        # padding
        if next:
            outstream.write(bytes(next.offset - outstream.tell()))



def _create_properties_table(font, format, base):
    """Create the Properties table."""
    propstrings = bytearray()
    xlfd_props = create_xlfd_properties(font)
    xlfd_props['FONT'] = create_xlfd_name(xlfd_props)
    props = []
    props_struct = base.Struct(**_PROPS)
    for key, value in xlfd_props.items():
        prop = props_struct(
            name_offset=len(propstrings),
            isStringProp=isinstance(value, str),
        )
        propstrings += key.encode('ascii', 'replace') + b'\0'
        if prop.isStringProp:
            prop.value = len(propstrings)
            value = from_quoted_string(value)
            propstrings += value.encode('ascii', 'replace') + b'\0'
        else:
            prop.value = int(value)
        props.append(prop)
    table_bytes = (
        bytes(le.uint32(format))
        + bytes(base.uint32(len(props)))
        + bytes((props_struct * len(props))(*props))
        # pad to next int32 boundary
        + bytes(0 if len(props)&3 == 0 else 4-(len(props)&3))
        + bytes(base.uint32(len(propstrings)))
        + bytes(propstrings)
    )
    return table_bytes, format


def _create_glyph_metrics(glyph, base):
    """Convert monobit glyph properties to PCF metrics."""
    metrics = base.Struct(**_UNCOMPRESSED_METRICS)(
        left_side_bearing=glyph.left_bearing,
        right_side_bearing=glyph.left_bearing + glyph.width,
        character_width=glyph.advance_width,
        character_ascent=glyph.height+glyph.shift_up,
        character_descent=-glyph.shift_up,
        character_attributes=0,
    )
    return metrics

def _create_ink_metrics(glyph, base):
    """Create the Ink Metrics table."""
    return _create_glyph_metrics(glyph.reduce(), base)


def _aggregate_metrics(metrics, aggfunc, base):
    """Get minimum or maximum from listo of metrics."""
    return base.Struct(**_UNCOMPRESSED_METRICS)(
        left_side_bearing=aggfunc(_m.left_side_bearing for _m in metrics),
        right_side_bearing=aggfunc(_m.right_side_bearing for _m in metrics),
        character_width=aggfunc(_m.character_width for _m in metrics),
        character_ascent=aggfunc(_m.character_ascent for _m in metrics),
        character_descent=aggfunc(_m.character_descent for _m in metrics),
        character_attributes=0,
    )

# xc/lib/font/util/fontaccel.c
# FontComputeInfoAccelerators(pFontInfo)
#     FontInfoPtr pFontInfo;
# {
#     pFontInfo->noOverlap = FALSE;
#     if (pFontInfo->maxOverlap <= pFontInfo->minbounds.leftSideBearing)
# 	pFontInfo->noOverlap = TRUE;
#
#     if ((pFontInfo->minbounds.ascent == pFontInfo->maxbounds.ascent) &&
# 	    (pFontInfo->minbounds.descent == pFontInfo->maxbounds.descent) &&
# 	    (pFontInfo->minbounds.leftSideBearing ==
# 	     pFontInfo->maxbounds.leftSideBearing) &&
# 	    (pFontInfo->minbounds.rightSideBearing ==
# 	     pFontInfo->maxbounds.rightSideBearing) &&
# 	    (pFontInfo->minbounds.characterWidth ==
# 	     pFontInfo->maxbounds.characterWidth) &&
#       (pFontInfo->minbounds.attributes == pFontInfo->maxbounds.attributes)) {
# 	pFontInfo->constantMetrics = TRUE;
# 	if ((pFontInfo->maxbounds.leftSideBearing == 0) &&
# 		(pFontInfo->maxbounds.rightSideBearing ==
# 		 pFontInfo->maxbounds.characterWidth) &&
# 		(pFontInfo->maxbounds.ascent == pFontInfo->fontAscent) &&
# 		(pFontInfo->maxbounds.descent == pFontInfo->fontDescent))
# 	    pFontInfo->terminalFont = TRUE;
# 	else
# 	    pFontInfo->terminalFont = FALSE;
#     } else {
# 	pFontInfo->constantMetrics = FALSE;
# 	pFontInfo->terminalFont = FALSE;
#     }
#     if (pFontInfo->minbounds.characterWidth == pFontInfo->maxbounds.characterWidth)
# 	pFontInfo->constantWidth = TRUE;
#     else
# 	pFontInfo->constantWidth = FALSE;
#
#     if ((pFontInfo->minbounds.leftSideBearing >= 0) &&
# 	    (pFontInfo->maxOverlap <= 0) &&
# 	    (pFontInfo->minbounds.ascent >= -pFontInfo->fontDescent) &&
# 	    (pFontInfo->maxbounds.ascent <= pFontInfo->fontAscent) &&
# 	    (-pFontInfo->minbounds.descent <= pFontInfo->fontAscent) &&
# 	    (pFontInfo->maxbounds.descent <= pFontInfo->fontDescent))
# 	pFontInfo->inkInside = TRUE;
#     else
# 	pFontInfo->inkInside = FALSE;
# }

def _create_acc_table(font, format, base, create_ink_bounds):
    """Create the Accelerators table."""
    if create_ink_bounds:
        format |= PCF_ACCEL_W_INKBOUNDS
    if not font.glyphs:
        raise ValueError('No glyphs in font.')
    metrics = tuple(_create_glyph_metrics(_g, base) for _g in font.glyphs)
    ink_metrics = tuple(_create_ink_metrics(_g, base) for _g in font.glyphs)
    minbounds = _aggregate_metrics(metrics, min, base)
    maxbounds = _aggregate_metrics(metrics, max, base)
    if create_ink_bounds:
        ink_minbounds = _aggregate_metrics(ink_metrics, min, base)
        ink_maxbounds = _aggregate_metrics(ink_metrics, max, base)
    # based on fontaccel.c code and FontForge's reverse engineered description
    # this is max(left_bearing + width - advance_width) == max(-right_bearing)
    maxOverlap = max(_m.right_side_bearing - _m.character_width for _m in metrics)
    # based on fontaccel.c
    constantMetrics = minbounds == maxbounds
    # checked against bdftopcf and pcf2bdf - these are bdf's FONT_ASCENT and FONT_DESCENT
    # which are kept only here and not in XLFD properties
    fontAscent = font.ascent
    fontDescent = font.descent
    acc_table = base.Struct(**_ACC_TABLE)(
        # /* if for all i, max(metrics[i].rightSideBearing - metrics[i].characterWidth) */
        # /*      <= minbounds.leftSideBearing */
        noOverlap=(maxOverlap <= minbounds.left_side_bearing),
        # /* Means the perchar field of the XFontStruct can be NULL */
        constantMetrics=constantMetrics,
        # /* constantMetrics true and forall characters: */
        # /*      the left side bearing==0 */
        # /*      the right side bearing== the character's width */
        # /*      the character's ascent==the font's ascent */
        # /*      the character's descent==the font's descent */
        terminalFont=(
            constantMetrics
            and (maxbounds.left_side_bearing == 0)
            and (maxbounds.right_side_bearing == maxbounds.character_width)
            and (maxbounds.character_ascent == fontAscent)
            and (maxbounds.character_descent == fontDescent)
        ),
        # /* monospace font like courier */
        constantWidth=(minbounds.character_width == maxbounds.character_width),
        # /* Means that all inked bits are within the rectangle with x between [0,charwidth] */
        # /*  and y between [-descent,ascent]. So no ink overlaps another char when drawing */
        # ===> this seems to imply fontAscent + fontDescent includes leading
        inkInside=(
            (minbounds.left_side_bearing >= 0)
            and (maxOverlap <= 0)
            and (minbounds.character_ascent >= -fontDescent)
            and (maxbounds.character_ascent <= fontAscent)
            and (-minbounds.character_descent <= fontAscent)
            and (maxbounds.character_descent <= fontDescent)
        ),
        # /* true if the ink metrics differ from the metrics somewhere */
        inkMetrics=create_ink_bounds and any(_m != _i for _m, _i in zip(metrics, ink_metrics)),
        # /* 0=>left to right, 1=>right to left */
        # however in practice this is set to 0 even on fonts with RTL glyphs
        # e.g. /usr/share/fonts/arabic24.pcf.gz
        drawDirection=0,
        padding=0,
        fontAscent=fontAscent,
        fontDescent=fontDescent,
        # where set in X11 code?
        maxOverlap=maxOverlap,
        # minbounds=_UNCOMPRESSED_METRICS,
        # maxbounds=_UNCOMPRESSED_METRICS,
    )
    table_bytes = [
        bytes(le.uint32(format)),
        bytes(acc_table),
        bytes(minbounds),
        bytes(maxbounds),
    ]
    if create_ink_bounds:
        table_bytes.extend([
            bytes(ink_minbounds),
            bytes(ink_maxbounds),
    ])
    return b''.join(table_bytes), format


def _create_metrics_table(font, format, base, create_glyph_metrics):
    """Create the Metrics table."""
    # we don't set PCF_COMPRESSED_METRICS
    metrics = tuple(create_glyph_metrics(_g, base) for _g in font.glyphs)
    table_bytes = (
        bytes(le.uint32(format))
        + bytes(base.int32(len(metrics)))
        + b''.join(bytes(_t) for _t in metrics)
    )
    return table_bytes, format


def _create_bitmaps(
        font, format, base,
        scan_unit_bytes=1, padding_bytes=1, bit_order='little',
    ):
    """Create the Bitmaps table."""
    byte_big = base == be
    bit_big = bit_order[:1].lower() == 'b'
    bitmaps = (
        _g.as_bytes(
            # align rows on padding_bytes boundaries
            stride=ceildiv(_g.width, padding_bytes*8) * padding_bytes*8,
            byte_swap=0 if (bool(byte_big) == bool(bit_big)) else scan_unit_bytes,
            bit_order='big' if bit_big else 'little',
        )
        for _g in font.glyphs
    )
    # align full byte sequence on scan_unit boundaries
    bitmaps = tuple(
        _bits.ljust(ceildiv(len(_bits), scan_unit_bytes) * scan_unit_bytes)
        for _bits in bitmaps
    )
    offsets = tuple(accumulate((len(_b) for _b in bitmaps), initial=0))[:-1]
    offsets = (base.int32 * len(bitmaps))(*offsets)
    bitmap_data = b''.join(bitmaps)
    # bytes # shorts # ints #?
    # apparently we do need to calculate all 4
    bitmap_sizes = [
        sum(
            # align full byte sequence on scan_unit boundaries
            ceildiv(
                # align rows on padding_bytes boundaries
                _g.pixels.get_byte_size(stride=ceildiv(_g.width, 8*2**_p) * 8*(2**_p)),
                scan_unit_bytes
            ) * scan_unit_bytes
            for _g in font.glyphs
        )
        for _p in range(4)
    ]
    assert bitmap_sizes[format&3] == len(bitmap_data), f'{bitmap_sizes[format&3]} != {len(bitmap_data)}'
    bitmap_sizes = (base.int32 * 4)(*bitmap_sizes)
    table_bytes = (
        bytes(le.uint32(format))
        + bytes(base.int32(len(offsets)))
        + bytes(offsets)
        + bytes(bitmap_sizes)
        + bitmap_data
    )
    return table_bytes, format


def _create_encoding(font, format, base):
    """Create the Encoding table."""
    font = font.label(codepoint_from=font.encoding)
    enc = base.Struct(**_ENCODING_TABLE)(
        default_char=int(font.get_default_glyph().codepoint or 0)
    )
    codepoints = font.get_codepoints()
    if not codepoints:
        raise ValueError('No storable code points in font.')
    byte_length = len(max(codepoints))
    if byte_length > 2:
        logging.warning(
            'Code points greater than 2 bytes cannot be stored in PCF.'
        )
    elif byte_length == 1:
        enc.min_byte1 = enc.max_byte1 = 0
        enc.min_char_or_byte2 = ord(min(codepoints))
        enc.max_char_or_byte2 = ord(max(codepoints))
    elif byte_length == 2:
        byte2 = [_cp[0] for _cp in codepoints if len(_cp) == 2]
        if any(len(_cp) == 1 for _cp in codepoints):
            byte2.append(0)
        enc.min_char_or_byte2 = min(byte2)
        enc.max_char_or_byte2 = max(byte2)
        byte1 = tuple(_cp[-1] for _cp in codepoints)
        enc.min_byte1 = min(byte1)
        enc.max_byte1 = max(byte1)
    glyph_indices = []
    for cp in _generate_codepoints(enc):
        try:
            index = font.get_index(cp)
        except KeyError:
            # -1 means 'not used'
            index = -1
        glyph_indices.append(index)
    glyph_indices = (base.int16 * len(glyph_indices))(*glyph_indices)
    table_bytes = (
        bytes(le.uint32(format))
        + bytes(enc)
        + bytes(glyph_indices)
    )
    return table_bytes, format


def _create_swidths(font, format, base):
    """Create the Scalable Widths table."""
    swidths = (base.int32 * len(font.glyphs))(*(
        pixel_to_swidth(_g.scalable_width, font.point_size, font.dpi.x)
        for _g in font.glyphs
    ))
    table_bytes = (
        bytes(le.uint32(format))
        + bytes(base.uint32(len(swidths)))
        + bytes(swidths)
    )
    return table_bytes, format


def _create_glyph_names(font, format, base):
    """Create the Glyph Names table."""
    names = tuple(
        _g.tags[0].value.encode('ascii', 'replace') + b'\0'
        if _g.tags else b'glyph%d\0' % (_i,)
        for _i, _g in enumerate(font.glyphs)
    )
    strings = b''.join(_n for _n in names)
    offsets = tuple(accumulate((len(_n) for _n in names), initial=0))[:-1]
    offsets = (base.uint32 * len(names))(*offsets)
    table_bytes = (
        bytes(le.uint32(format))
        + bytes(base.uint32(len(offsets)))
        + bytes(offsets)
        + bytes(base.uint32(len(strings)))
        + strings
    )
    return table_bytes, format



"""
monobit.blocks - output pixels as text using block elements

(c) 2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import zip_longest

from .binary import bytes_to_bits


class blockstr(str):
    """str that is shown as block text in interactive session."""
    def __repr__(self):
        return f'"""\\\n{self}"""'


# block elements
# this has nothing to do with `blockstr` above

BLOCKS = {
    (1, 1): {
        (0,): ' ',
        (1,): '\u2588',
    },
    (2, 2): {
        (0, 0, 0, 0): ' ',
        (0, 0, 0, 1): '\u2597',
        (0, 0, 1, 0): '\u2596',
        (0, 0, 1, 1): '\u2584',
        (0, 1, 0, 0): '\u259d',
        (0, 1, 0, 1): '\u2590',
        (0, 1, 1, 0): '\u259e',
        (0, 1, 1, 1): '\u259f',
        (1, 0, 0, 0): '\u2598',
        (1, 0, 0, 1): '\u259a',
        (1, 0, 1, 0): '\u258c',
        (1, 0, 1, 1): '\u2599',
        (1, 1, 0, 0): '\u2580',
        (1, 1, 0, 1): '\u259c',
        (1, 1, 1, 0): '\u259b',
        (1, 1, 1, 1): '\u2588',
    },
}

# sixel block elements in Unicode do not include full- and half-block elements defined elsewhere
_SIXBITS = tuple(
    tuple(1*_b for _b in bytes_to_bits((_code,))[2:])
    for _code in range(1, 2**6-1)
    if _code not in (0b010101, 0b101010)
)
BLOCKS[(2, 3)] = {
    (_0, _1, _2, _3, _4, _5): chr(0x1Fb00 + _i)
    for _i, (_5, _4, _3, _2, _1, _0) in enumerate(_SIXBITS)
} | {
    (0, 0, 0, 0, 0, 0): ' ',
    (0, 1, 0, 1, 0, 1): '\u2590',
    (1, 0, 1, 0, 1, 0): '\u258c',
    (1, 1, 1, 1, 1, 1): '\u2588',
}

_EIGHTBITS = tuple(
    tuple(1*_b for _b in bytes_to_bits((_code,)))
    for _code in range(2**8)
)
BLOCKS[(2, 4)] = {
    (_0, _3, _1, _4, _2, _5, _6, _7): chr(0x2800 + _i)
    for _i, (_7, _6, _5, _4, _3, _2, _1, _0) in enumerate(_EIGHTBITS)
}

BLOCKS[(2, 1)] = {
    (_0, _1): BLOCKS[(2, 2)][(_0, _1, _0, _1)]
    for _0 in range(2)
    for _1 in range(2)
}

BLOCKS[(1, 2)] = {
    (_0, _1): BLOCKS[(2, 2)][(_0, _0, _1, _1)]
    for _0 in range(2)
    for _1 in range(2)
}

BLOCKS[(1, 3)] = {
    (_0, _1, _2): BLOCKS[(2, 3)][(_0, _0, _1, _1, _2, _2)]
    for _0 in range(2)
    for _1 in range(2)
    for _2 in range(2)
}


def matrix_to_blocks(matrix, ncols, nrows):
    """Convert bit matrix to a string of block characters."""
    try:
        blockdict = BLOCKS[(ncols, nrows)]
    except KeyError:
        raise ValueError(f'Unsupported block resolution: {ncols}x{nrows}')
    bitblockrows = tuple(
        tuple(
            _bitblock
            for _bitblock in zip_longest(
                *(
                    _bitrows[_row][_col::ncols]
                    for _row in range(nrows)
                    for _col in range(ncols)
                ),
                fillvalue=0
            )
        )
        for _bitrows in zip_longest(
            *(matrix[_ofs::nrows] for _ofs in range(nrows)), fillvalue=()
        )
    )
    blocks = '\n'.join(
        ''.join(blockdict[_bitblock] for _bitblock in _row)
        for _row in bitblockrows
    )
    return blockstr(blocks + '\n')



"""
monobit.formats.text.draw - visual-text formats

(c) 2019--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
import string

from ...storage import loaders, savers
from ...font import Font
from ...glyph import Glyph
from ...labels import Tag, Char
from ...magic import FileFormatError
from ...encoding import charmaps
from ...binary import align


##############################################################################
# hexdraw

@loaders.register(
    name='hexdraw',
    patterns=('*.draw',),
)
def load_hexdraw(instream, ink:str='#', paper:str='-', unicode:bool=True):
    """
    Load font from a hexdraw file.

    ink: character used for inked/foreground pixels (default #)
    paper: character used for uninked/background pixels (default -)
    unicode: interpret codepoint as Unicode (default: True)
    """
    DrawGlyph.ink = ink
    DrawGlyph.paper = paper
    return load_draw(
        instream.text, blocktypes=(DrawGlyph, DrawComment, Empty),
        unicode=unicode
    )


@savers.register(linked=load_hexdraw)
def save_hexdraw(fonts, outstream, ink:str='#', paper:str='-', unicode:bool=True):
    """
    Save font to a hexdraw file.

    ink: character to use for inked/foreground pixels (default #)
    paper: character to use for uninked/background pixels (default -)
    unicode: use unicode char labels for codepoints (default: True)
    """
    if len(fonts) > 1:
        raise FileFormatError("Can only save one font to hexdraw file.")
    _save_draw(fonts[0], outstream.text, ink=ink, paper=paper, unicode=unicode)


# read hexdraw file

def load_draw(text_stream, *, blocktypes, unicode):
    """Parse a hexdraw-style file."""
    glyphs = []
    font_comments = []
    current_comment = []
    for block in iter_blocks(text_stream, blocktypes):
        if isinstance(block, DrawComment):
            if not glyphs:
                font_comments.extend(current_comment)
                current_comment = []
            current_comment.append(block.get_value())
        elif isinstance(block, DrawGlyph):
            if not glyphs and not font_comments:
                font_comments.extend(current_comment)
                current_comment = []
            glyphs.append(block.get_value().modify(
                comment='\n\n'.join(current_comment),
            ))
            current_comment = []
        elif isinstance(block, Unparsed):
            logging.debug('Unparsed lines: %s', block.get_value())
    font_comments.extend(current_comment)
    if not unicode:
        glyphs = tuple(
            _g.label(codepoint_from=charmaps['unicode']).modify(char=None)
            for _g in glyphs
        )
    font = Font(
        glyphs,
        comment='\n\n'.join(font_comments)
    )

    return font


# write hexdraw file

def _save_draw(font, outstream, *, ink, paper, unicode):
    """Write one font to a plaintext stream as hexdraw."""
    font = font.equalise_horizontal()
    if unicode:
        # ensure char labels are set
        font = font.label(char_from=font.encoding, match_whitespace=False, match_graphical=False)
    # write global comment
    if font.get_comment():
        outstream.write(
            format_comment(font.get_comment(), comment_char='%') + '\n'
        )
    # write glyphs
    for i, glyph in enumerate(font.glyphs):
        if unicode:
            char = glyph.char
        else:
            cpbytes = bytes(glyph.codepoint)
            cpbytes = cpbytes.rjust(align(len(cpbytes), 2), b'\0')
            char = cpbytes.decode('utf-32-be')
        if not char:
            logging.warning(
                "Can't encode glyph without Unicode character label in .draw file;"
                " skipping index %d", i
            )
        elif len(char) > 1:
            logging.warning(
                "Can't encode grapheme cluster %s in .draw file; skipping.",
                ascii(char)
            )
        else:
            glyphtxt = glyph.as_text(start='\t', ink=ink, paper=paper, end='\n')
            outstream.write(f'\n{ord(char):04x}:')
            outstream.write(glyphtxt)


def format_comment(comments, comment_char):
    """Format a multiline comment."""
    return '\n'.join(
        f'{comment_char} {_line}'
        for _line in comments.splitlines()
    )


###############################################################################
# mkwinfon .fd

from ...properties import Props
from ..windows import WEIGHT_MAP, CHARSET_MAP


FD_KEYS = {
    'facename',
    'copyright',
    'height',
    'ascent',
    'pointsize',
    'weight',
    'charset',
    'italic',
    'underline',
    'strikeout',
}

FD_CHAR_KEYS = {
    'char',
    'width',
}

@loaders.register(
    name='mkwinfon',
    patterns=('*.fd',),
)
def load_mkwinfon(instream):
    """Load font from a mkwinfon .fd file."""
    properties, glyphs, comments = _read_mkwinfon(instream.text)
    return _convert_mkwinfon(properties, glyphs, comments)


def _read_mkwinfon(text_stream):
    """Read a mkwinfon file into a properties object."""
    # this will be a list of tuples
    font_props = {_k: None for _k in FD_KEYS}
    font_comments = []
    glyphs = []
    glyph_props = {}
    current_comment = []
    for block in iter_blocks(text_stream, (MWFGlyph, MWFProperties, MWFComment, Empty)):
        if isinstance(block, MWFComment):
            if not glyphs:
                font_comments.extend(current_comment)
                current_comment = []
            current_comment.append(block.get_value())
        elif isinstance(block, MWFProperties):
            properties = block.get_value()
            if glyphs or 'char' in properties:
                glyph_props.update(properties)
            else:
                font_props.update(properties)
                if current_comment:
                    font_comments.extend(current_comment)
                    current_comment = []
        elif isinstance(block, MWFGlyph):
            if not glyphs and not font_comments:
                font_comments.extend(current_comment)
                current_comment = []
            glyphs.append(Glyph(
                block.get_value(), _0='0', _1='1',
                codepoint=glyph_props.pop('char', b''),
                comment='\n\n'.join(current_comment),
            ))
            glyph_props = {}
            current_comment = []
        elif isinstance(block, Unparsed):
            logging.debug('Unparsed lines: %s', block.get_value())
    font_comments.extend(current_comment)
    return Props(**font_props), glyphs, font_comments

def _convert_mkwinfon(props, glyphs, comments):
    mb_props = dict(
        name=props.facename,
        copyright=props.copyright,
        descent=int(props.height)-int(props.ascent),
        ascent=props.ascent,
        point_size=props.pointsize,
        weight=(
            None if props.weight is None else
            WEIGHT_MAP.get(round(max(100, min(900, int(props.weight))), -2), None)
        ),
        encoding=(
            None if props.charset is None else
            CHARSET_MAP.get(int(props.charset), None)
        ),
        comment='\n\n'.join(comments),
    )
    return Font(glyphs, **mb_props).label()


###############################################################################
# consoleet / vfontas / hxtools

from pathlib import Path
from ...containers.directory import Directory

@loaders.register(
    name='consoleet',
)
def load_clt(instream):
    """Load font from consoleet files."""
    # this format consists of separate image files, without a manifest
    # instream.where does not give the nearest enclosing container but the root where we're calling!
    # we also can't use a directory as instream as it would be recursively read
    container = instream.where
    glyphs = []
    for name in sorted(container):
        if Path(name).parent != Path(instream.name).parent:
            continue
        with container.open(name, mode='r') as stream:
            glyphs.append(_read_clt_glyph(stream))
    return Font(glyphs, source_name=Path(instream.name).parent)

def _read_clt_glyph(instream):
    text = instream.text
    name = instream.name
    codepoint = Path(name).stem
    magic = text.readline().strip()
    if magic != 'PCLT':
        return Glyph()
    width, _, height = text.readline().strip().partition(' ')
    glyphtext = text.read().splitlines()
    return Glyph(
        glyphtext, _0='.', _1='#',
        # encoding is not specified by spec or file - can be unicode or codepage
        codepoint=f'0x{codepoint}'
    ).shrink(factor_x=2)


###############################################################################
# psf2txt

PSFT_KEYS = {
    'Version',
    'Flags',
    'Length',
    'Width',
    'Height',
}

PSFT_CHAR_KEYS = {
    'Unicode',
}

@loaders.register(
    name='psf2txt',
    magic=(b'%PSF2',),
    patterns=('*.txt',),
)
def load_psf2txt(instream):
    """Load font from a psf2txt .txt file."""
    properties, glyphs, comments = _read_psf2txt(instream.text)
    return _convert_psf2txt(properties, glyphs, comments)


def _read_psf2txt(text_stream):
    """Read a psf2txt file into a properties object."""
    if text_stream.readline().strip() != '%PSF2':
        raise FileFormatError('Not a PSF2TXT file.')
    properties = {_k: None for _k in PSFT_KEYS}
    comment = []
    glyphs = []
    current_comment = comment
    current_props = properties
    for block in iter_blocks(text_stream, (PTSeparator, PTComment, PTGlyph, PTLabel, PTProperties, Empty)):
        if isinstance(block, PTSeparator):
            if glyphs:
                glyphs[-1] = glyphs[-1].modify(
                    comment='\n\n'.join(current_comment),
                    **current_props
                )
            else:
                properties.update(current_props)
            current_comment = []
            current_props = {}
        elif isinstance(block, PTComment):
            current_comment.append(block.get_value())
        elif isinstance(block, PTProperties):
            current_props.update(block.get_value())
        elif isinstance(block, PTGlyph):
            glyphs.append(Glyph(block.get_value(), _0='-', _1='#'))
        elif isinstance(block, PTLabel):
            glyphs[-1] = glyphs[-1].modify(labels=block.get_value())
        elif isinstance(block, Unparsed):
            logging.debug('Unparsed lines: %s', block.get_value())
    if current_comment or current_props:
        glyphs[-1] = glyphs[-1].modify(
            comment='\n\n'.join(current_comment),
            **current_props
        )
    return Props(**properties), glyphs, comment

def _convert_psf2txt(props, glyphs, comment):
    mb_props = dict(
        revision=props.Version,
        # ignore Flags, we don't need the others
    )
    return Font(glyphs, **mb_props, comment='\n\n'.join(comment))


##############################################################################
# common utilities and reader classes

def iter_blocks(text_stream, classes):
    block = BlockBuilder(classes)
    for line in text_stream:
        while not block.append(line.rstrip('\r\n')):
            yield block.emit()
            block = BlockBuilder(classes)
    if block:
        yield block.emit()


class BlockBuilder:

    def __init__(self, classes):
        self.block = None
        self.classes = (*classes, Unparsed)

    def __bool__(self):
        return self.block is not None

    def append(self, line):
        if not self.block:
            for blocktype in self.classes:
                try:
                    self.block = blocktype(line)
                    return True
                except ValueError:
                    pass
            else:
                # this should not happen - BaseBlock absorbs
                raise ValueError('unparsed block')
        if self.block.ends(line):
            return False
        self.block.append(line)
        return True

    def emit(self):
        return self.block


class BaseBlock:

    def __init__(self, line):
        if not self.starts(line):
            raise ValueError(f'{line:r} cannot start a {type(self).__name__}')
        self.lines = []
        self.append(line)

    def __repr__(self):
        return f'{type(self).__name__}({repr(self.lines)})'

    def starts(self, line):
        return True

    def ends(self, line):
        return True

    def append(self, line):
        self.lines.append(line)

    def get_value(self):
        return tuple(self.lines)


class Unparsed(BaseBlock):
    pass


class Empty(BaseBlock):

    def starts(self, line):
        return not line


class NonEmptyBlock(BaseBlock):

    def append(self, line):
        line = line.rstrip()
        if line:
            self.lines.append(line)


# draw format block readers


class DrawComment(NonEmptyBlock):

    notcomment = string.hexdigits + string.whitespace

    def starts(self, line):
        return line[:1] not in self.notcomment

    def ends(self, line):
        return not self.starts(line)

    def get_value(self):
        lines = tuple(self.lines)
        first = equal_firsts(lines)
        if first and first not in string.ascii_letters + string.digits:
            lines = tuple(_line[1:] for _line in lines)
        if equal_firsts(lines) == ' ':
            lines = (_line[1:] for _line in lines)
        return '\n'.join(lines)

def equal_firsts(lines):
    first_chars = set(_line[:1] for _line in lines if _line)
    if len(first_chars) == 1:
        return first_chars.pop()
    return None


class DrawGlyph(NonEmptyBlock):
    paper = '-'
    ink = '#'

    def starts(self, line):
        return line and line[:1] in string.hexdigits

    def ends(self, line):
        # empty or non-indented
        return not line[:1] in string.whitespace

    def get_value(self):
        key, _, value =  self.lines[0].partition(':')
        value = value.strip()
        lines = self.lines[1:]
        if value:
            lines = [value] + lines
        lines = tuple(_l.strip() for _l in lines)
        return Glyph(
            lines, _0=self.paper, _1=self.ink,
            labels=(self.convert_key(key),),
        )

    @staticmethod
    def convert_key(key):
        """Convert keys on input from .draw."""
        key = key.strip()
        try:
            return Char(''.join(chr(int(_key, 16)) for _key in key.split(',')))
        except (TypeError, ValueError):
            return Tag(key)


# mkwinfont block readers

class MWFGlyph(NonEmptyBlock):

    def starts(self, line):
        return line[:1] in ('0', '1')

    def ends(self, line):
        return line[:1] not in ('0', '1')

    def get_value(self):
        return tuple(_l.strip() for _l in self.lines)


class MWFProperties(NonEmptyBlock):

    def starts(self, line):
        return line[:1] in string.ascii_letters and ' ' in line

    def ends(self, line):
        return not self.starts(line)

    def get_value(self):
        return dict(_l.split(' ', 1) for _l in self.lines)

class MWFComment(DrawComment):
    pass


# psf2txt block readers

class PTSeparator(NonEmptyBlock):

    def starts(self, line):
        return line[:1] == '%'


class PTComment(NonEmptyBlock):

    def starts(self, line):
        return line.startswith('//')

    def ends(self, line):
        return not self.starts(line)

    def get_value(self):
        lines = tuple(_l.removeprefix('//') for _l in self.lines)
        if equal_firsts(lines) == ' ':
            lines = (_line[1:] for _line in lines)
        return '\n'.join(lines)


class PTGlyph(NonEmptyBlock):

    def starts(self, line):
        return line.startswith('Bitmap:')

    def ends(self, line):
        return line[:1] not in string.whitespace

    def append(self, line):
        line = line.rstrip('\\').strip()
        if line:
            self.lines.append(line)

    def get_value(self):
        _, _, value =  self.lines[0].partition(':')
        value = value.strip()
        lines = self.lines[1:]
        if value:
            lines = [value] + lines
        lines = tuple(_l.strip() for _l in lines)
        return lines


class PTProperties(NonEmptyBlock):

    def starts(self, line):
        return line[:1] in string.ascii_letters and ':' in line

    def ends(self, line):
        return not self.starts(line)

    def get_value(self):
        return dict((_e.strip() for _e in _l.split(':', 1)) for _l in self.lines)


class PTLabel(NonEmptyBlock):

    def starts(self, line):
        return line.startswith('Unicode:')

    def get_value(self):
        _, _, value =  self.lines[0].partition(':')
        value = value.strip()[1:-2].split('];[')
        return tuple(
            Char(''.join(
                chr(int(_cp, 16)) for _cp in _l.split('+'))
            )
            for _l in value
        )
